# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-12-05 12:01:21
# @Last modified by: José Sánchez-Gallego
# @Last Modified time: 2017-12-05 12:19:32

from __future__ import annotations

import inspect

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from jaeger.commands.trajectory import Trajectory


class JaegerError(Exception):
    """A custom core Jaeger exception"""

    def __init__(self, message=None):
        message = "There has been an error" if not message else message
        super(JaegerError, self).__init__(message)


class FPSLockedError(JaegerError):
    """The FPS is locked."""


class JaegerNotImplemented(JaegerError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):
        message = "This feature is not implemented yet." if not message else message
        super(JaegerNotImplemented, self).__init__(message)


class JaegerCANError(JaegerError):
    """Exception class for CAN-related errors."""

    def __init__(self, message=None, serial_reply=None):

        if message is None:
            message = ""

        if serial_reply is not None:
            message = message.strip() + " " + serial_reply

        super(JaegerCANError, self).__init__(message)


class PositionerError(JaegerError):
    """Exception class for positioner-related errors."""

    def __init__(self, message=None, positioner=None):

        if message is None:
            message = ""

        if positioner is not None:
            pid = positioner.positioner_id
        else:
            stack = inspect.stack()
            f_locals = stack[1][0].f_locals

            if "self" in f_locals:
                pid = f_locals["self"].positioner_id
            else:
                pid = "UNKNOWN"

        message = f"Positioner {pid}: {message}"

        super(PositionerError, self).__init__(message)


class CommandError(JaegerError):
    """Exception class for command-related errors."""

    def __init__(self, message=None, command=None):

        if message is None:
            message = ""

        if command is None:
            stack = inspect.stack()
            f_locals = stack[1][0].f_locals

            if "self" in f_locals:
                command = f_locals["self"]

        if command:
            c_name = command.name
            c_uid = command.command_uid
            message = f"({c_name}, {c_uid!s}): {message}"

        super(CommandError, self).__init__(message)


class JaegerMissingDependency(JaegerError):
    """A custom exception for missing dependencies."""

    pass


class TrajectoryError(JaegerError):
    """A trajectory error."""

    def __init__(self, message=None, trajectory: Trajectory | None = None):
        if message and isinstance(message, str) and message[-1] == ".":
            message = message[:-1]

        super().__init__(message)

        self.trajectory = trajectory
        if self.trajectory:
            self.trajectory.failed = True


class FVCError(JaegerError):
    """An error handling the FVC or the FVC loop."""

    pass


class JaegerWarning(Warning):
    """Base warning for Jaeger."""

    pass


class JaegerUserWarning(UserWarning, JaegerWarning):
    """The primary warning class."""

    pass


class JaegerSkippedTestWarning(JaegerUserWarning):
    """A warning for when a test is skipped."""

    pass


class JaegerDeprecationWarning(JaegerUserWarning):
    """A warning for deprecated features."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-26
# @Filename: alerts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
from contextlib import suppress

from typing import TYPE_CHECKING, Any

from clu.legacy.tron import TronKey
from drift import Device, Relay

from jaeger import config
from jaeger.ieb import IEB, Chiller
from jaeger.utils.helpers import BaseBot


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = ["AlertsBot"]


class AlertsBot(BaseBot):
    """Monitors values and raises alerts."""

    def __init__(self, fps: FPS):

        super().__init__(fps)

        self.config: dict[str, Any] = config["alerts"]
        self.interval: float = self.config["interval"]

        self.keywords: dict[str, bool] = {}
        self._gfa_alerts: dict[str, bool] = {}

        self.reset()

    def reset(self):
        """Resets alerts and parameters."""

        self.keywords = {
            "alert_gfa_temp_critical": False,
            "alert_gfa_temp_warning": False,
            "alert_ieb_temp_critical": False,
            "alert_ieb_temp_warning": False,
            "alert_robot_temp_critical": False,
            "alert_robot_temp_warning": False,
            "alert_fps_flow": False,
            "alert_dew_point": False,
            "alert_chiller_dew_point": False,
            "alert_chiller_fault": False,
            "alert_fluid_temperature": False,
        }
        self._gfa_alerts = {}

    async def start(self, delay: float | bool = False):
        """Stars the monitoring loop."""

        await self.stop()

        if delay is not False and delay > 0:
            await asyncio.sleep(delay)

        self._task = asyncio.create_task(self._loop())

        if "gfa" in self.config["enabled"]:
            if self.actor and "fliswarm" in self.actor.models:
                self.actor.models["fliswarm"].register_callback(self._check_gfa)
            else:
                self.notify(
                    "Failed starting GFA alert monitoring.",
                    level=logging.ERROR,
                )

    async def stop(self):
        """Stops the monitoring loop."""

        if self._task:
            with suppress(asyncio.CancelledError):
                self._task.cancel()
                await self._task
        if (
            self.actor
            and "fliswarm" in self.actor.models
            and self._check_gfa in self.actor.models["fliswarm"]._callbacks
        ):
            self.actor.models["fliswarm"].remove_callback(self._check_gfa)

    def set_keyword(self, keyword: str, new_value: bool) -> bool:
        """Sets the value of an alert keyword and outputs it to the actor.

        Returns a boolean indicating whether the value has changed.
        """

        if keyword not in self.keywords:
            raise KeyError(f"Invalid alert keyword {keyword}.")

        changed = self.keywords[keyword] != new_value

        self.keywords[keyword] = new_value

        if new_value is True:
            level = logging.WARNING
        else:
            level = logging.INFO

        # Repeatedly output the keyword if the alert is on.
        # Otherwise only if it changed.
        if new_value is True or changed is True:
            self.notify({keyword: int(self.keywords[keyword])}, level=level)

        return changed

    async def _loop(self):
        """The main monitoring loop."""

        coros = []
        if "robot" in self.config["enabled"]:
            coros.append(self._check_robots)
        if "ieb" in self.config["enabled"]:
            coros.append(self._check_ieb)
        if "flow" in self.config["enabled"]:
            coros.append(self._check_flow)
        if "temperature" in self.config["enabled"]:
            coros.append(self._check_outside_temperature)
        if "chiller" in self.config["enabled"]:
            coros.append(self._check_chiller)

        while True:

            for coro in coros:
                try:
                    await coro()
                except Exception as err:
                    self.notify(
                        f"Failed running alerts coroutine {coro.__name__}: {err}"
                    )

            await asyncio.sleep(self.interval)

    async def get_dew_point_temperarure(self):
        """Returns the ambient and dew point temperatures."""

        assert isinstance(self.ieb, IEB)

        temp_config = config["alerts"]["temperature"]

        temp = (await self.ieb.read_device(temp_config["sensor_temp"], adapt=True))[0]
        rh = (await self.ieb.read_device(temp_config["sensor_rh"], adapt=True))[0]

        # Dewpoint temperature.
        t_d = temp - (100 - rh) / 5.0

        return temp, t_d

    async def shutdown_gfas(self):
        """Shutdowns the GFAs without touching the rest of the FPS."""

        if not isinstance(self.ieb, IEB):
            self.notify(
                "IEB not connected, cannot power off GFAs.",
                level=logging.ERROR,
            )
            return

        self.notify("Shutting down cameras.")

        for gfa in range(1, 7):
            device = self.ieb.get_device(f"GFA{gfa}")

            assert isinstance(device, Relay)
            await device.open()

            await asyncio.sleep(0.5)

    async def _shutdown_device(self, device: Device | str):
        """Shuts down a device."""

        if isinstance(device, str):
            if isinstance(self.ieb, IEB):
                device = self.ieb.get_device(device)
            else:
                self.notify(
                    f"IEB not connected, cannot find device {device}.",
                    level=logging.ERROR,
                )
                return

        assert isinstance(device, Relay)
        await device.open()

    async def shutdown_fps(
        self,
        nucs: bool = False,
        gfas: bool = False,
        cans: bool = False,
    ):
        """Shutdowns the robots and optionally other electronics."""

        if not isinstance(self.ieb, IEB):
            self.notify(
                "IEB not connected, cannot power off FPS.",
                level=logging.ERROR,
            )
            return

        self.notify("Shutting down power supplies.")
        for ps in range(1, 7):
            await self._shutdown_device(f"PS{ps}")

        if gfas is True:
            await self.shutdown_gfas()

        if cans is True:
            self.notify("Shutting down CAN devices.")
            for can in range(1, 7):
                await self._shutdown_device(f"CM{can}")

        if nucs is True:
            self.notify("Shutting down NUCs.")
            for nuc in range(1, 7):
                await self._shutdown_device(f"NUC{nuc}")

    async def _check_robots(self):
        """Checks robot temperature."""

        if not isinstance(self.ieb, IEB):
            self.notify("IEB not connected. Cannot check robot temperatures.")
            return

        robot_config = config["alerts"]["robot"]

        sensor = robot_config["sensor"]
        temperature = (await self.ieb.read_device(sensor, adapt=True))[0]

        if temperature > robot_config["critical"]:
            changed = self.set_keyword("alert_robot_temp_critical", True)
            if not changed:
                return

            self.notify("Critical robot temperature reached.")
            await self.shutdown_fps()

        elif temperature >= robot_config["warning"]:
            self.set_keyword("alert_robot_temp_warning", True)
            self.notify("Robot temperature exceeds safe limits.")

        else:
            self.set_keyword("alert_robot_temp_critical", False)
            self.set_keyword("alert_robot_temp_warning", False)

    async def _check_ieb(self):
        """Checks IEB internal temperature."""

        if not isinstance(self.ieb, IEB):
            self.notify("IEB not connected. Cannot check IEB temperature.")
            return

        ieb_config = config["alerts"]["ieb"]

        sensor = ieb_config["sensor"]
        temperature = (await self.ieb.read_device(sensor, adapt=True))[0]

        if temperature > ieb_config["critical"]:
            changed = self.set_keyword("alert_ieb_temp_critical", True)
            if not changed:
                return

            self.notify("Critical IEB temperature reached.")
            await self.shutdown_fps(nucs=True, gfas=True, cans=True)

        elif temperature >= ieb_config["warning"]:
            self.set_keyword("alert_ieb_temp_warning", True)
            self.notify("IEB temperature exceeds safe limits.")

        else:
            self.set_keyword("alert_ieb_temp_critical", False)
            self.set_keyword("alert_ieb_temp_warning", False)

        # If a GFA has caused a temperature alert and it's then disconnected
        # the alert won't clear because that camera stops reporting status.
        # To prevent that here we loop over the power status of each camera
        # and if it's off we disable the alert. This does not immediately disable
        # the alarm but next time that _check_gfa() is called it will refresh
        # the keywords.
        for gfa_id in range(1, 7):
            relay_status = await self.ieb.read_device(f"GFA{gfa_id}")
            if relay_status == "open":
                self._gfa_alerts.pop(f"gfa{gfa_id}", None)

    async def _check_gfa(self, model: dict, key: TronKey):
        """Check GFA temperatures."""

        if key.name != "status":
            return

        gfa_config = config["alerts"]["gfa"]

        camera_name: str = key.value[0]
        if not camera_name.startswith("gfa"):
            return

        base_temperature: float = float(key.value[17])

        if base_temperature >= gfa_config["critical"]:
            changed = self.set_keyword("alert_gfa_temp_critical", True)
            if not changed:
                return

            self.notify(f"Critical GFA temperature reached on camera {camera_name}.")
            self._gfa_alerts[camera_name] = True

            # This will only run once since once we shut down the GFAs the keyword
            # is not output anymore.
            await self.shutdown_gfas()

        elif base_temperature >= gfa_config["warning"]:
            self.set_keyword("alert_gfa_temp_warning", True)
            self.notify(f"GFA {camera_name} temperature exceeds safe limits.")
            self._gfa_alerts[camera_name] = True

        else:
            self._gfa_alerts[camera_name] = False
            if all([value is False for value in self._gfa_alerts.values()]):
                self.set_keyword("alert_gfa_temp_critical", False)
                self.set_keyword("alert_gfa_temp_warning", False)

    async def _check_flow(self):
        """Check flow rates."""

        if not isinstance(self.ieb, IEB):
            self.notify("IEB not connected. Cannot check flow rates.")
            return

        flow_config = config["alerts"]["flow"]

        sensor = flow_config["sensor"]
        flow = (await self.ieb.read_device(sensor, adapt=True))[0]

        if flow < flow_config["critical"]:
            self.set_keyword("alert_fps_flow", True)
            self.notify("FPS coolant flow is below limits.")

        else:
            self.set_keyword("alert_fps_flow", False)

    async def _check_outside_temperature(self):
        """Checks if the outside temperature is close to the dew point."""

        if not isinstance(self.ieb, IEB):
            self.notify("IEB not connected. Cannot check outside temperature.")
            return

        temp, t_d = await self.get_dew_point_temperarure()

        if temp < t_d + config["alerts"]["temperature"]["dew_threshold"]:
            self.set_keyword("alert_dew_point", True)
            self.notify("Outside temperature is approaching dew point limit.")

        else:
            self.set_keyword("alert_dew_point", False)

    async def _check_chiller(self):
        """Checks the chiller status."""

        if not isinstance(self.ieb, IEB):
            self.notify("IEB not connected. Cannot run chiller checks.")
            return

        chiller = Chiller.create()
        assert chiller is not None

        try:
            # setpoint = (await chiller.read_device("TEMPERATURE_USER_SETPOINT"))[0]
            fluid_temp = (await chiller.read_device("DISPLAY_VALUE"))[0]
        except Exception:
            # self.notify(f"Failed reading chiller values: {err}", level=logging.ERROR)
            return

        _, t_d = await self.get_dew_point_temperarure()

        if fluid_temp < t_d + config["alerts"]["temperature"]["dew_threshold"]:
            self.set_keyword("alert_chiller_dew_point", True)
            self.notify("Fluid temperature is approaching dew point limit.")

        else:
            self.set_keyword("alert_chiller_dew_point", False)

        # chiller_config = config["alerts"]["chiller"]

        # supply_temp = (await self.ieb.read_device(chiller_config["sensor_supply"]))[0]

        # if abs(setpoint - supply_temp) > chiller_config["threshold"]:
        #     self.set_keyword("alert_fluid_temperature", True)
        #     self.notify("Chiller set point is different from supply temperature.")

        # else:
        #     self.set_keyword("alert_fluid_temperature", False)

        # Check if there are chiller alerts.
        chiller_alerts: list[str] = []
        chiller_mod = chiller.modules["chiller"]
        for chiller_dev_name in chiller_mod.devices:
            if chiller_dev_name.startswith("alert_"):
                value: Any = await chiller.read_device(chiller_dev_name, adapt=False)
                if value > 0:
                    chiller_alerts.append(chiller_dev_name)

        if len(chiller_alerts) > 0:
            self.set_keyword("alert_chiller_fault", True)
            self.notify(
                "The following chiller alerts are active: " + ", ".join(chiller_alerts)
            )
        else:
            self.set_keyword("alert_chiller_fault", False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-22
# @Filename: debug.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from jaeger import __version__, config

from . import jaeger_parser


__all__ = ["debug"]


@jaeger_parser.group(invoke_without_command=True)
def debug():
    """Debug and engineering tools."""

    return


@debug.command("info")
def info_(command, fps):
    """Reports information about the system."""

    command.info(
        {
            "version": __version__,
            "config_file": config.CONFIG_FILE or "internal",
        },
        concatenate=False,
    )

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-27
# @Filename: chiller.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from . import JaegerCommandType, command_parser


if TYPE_CHECKING:
    from jaeger import FPS


__all__ = ["alerts"]


@command_parser.group()
def alerts(*args):
    """Manages the alerts system."""

    pass


@alerts.command()
async def status(command: JaegerCommandType, fps: FPS):
    """Shows the status of the alerts."""

    return command.finish(
        message={k: int(v) for k, v in command.actor.alerts.keywords.items()}
    )


@alerts.command()
async def reset(command: JaegerCommandType, fps: FPS):
    """Clears all the alerts. Needed after certain alerts have been raised."""

    command.actor.alerts.reset()
    return command.finish(
        message={k: int(v) for k, v in command.actor.alerts.keywords.items()}
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-11-12
# @Filename: testing.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import zlib

from typing import Optional, Tuple

import jaeger
from jaeger import config, utils
from jaeger.commands import CommandID
from jaeger.interfaces import Message, VirtualBus
from jaeger.maskbits import BootloaderStatus, PositionerStatus, ResponseCode
from jaeger.utils.helpers import StatusMixIn


__all__ = ["VirtualFPS", "VirtualPositioner"]


TIME_STEP = config["positioner"]["time_step"]


class VirtualFPS(jaeger.FPS):
    """A mock Focal Plane System for testing and development.

    This class listens to a virtual bus and responds as if real positioners were
    plugged into the system.

    """

    def __post_init__(self):

        self.can = "virtual"
        self.ieb = True
        self.observatory = "APO"

        super().__post_init__()

        self._vpositioner_bus = VirtualBus(config["profiles"]["virtual"]["channel"])
        self._vpositioners = {}

        asyncio.create_task(self.process_messages())

    def add_virtual_positioner(self, pid: int):

        self._vpositioners[pid] = VirtualPositioner(pid, bus=self._vpositioner_bus)

    def _check_fibre_assignments(self, *args, **kwargs):
        pass

    async def process_messages(self):

        while True:

            msg = await self._vpositioner_bus.get()

            if msg is None:
                continue

            arbitration_id = msg.arbitration_id
            positioner_id, command_id, uid, __ = utils.parse_identifier(arbitration_id)

            if positioner_id != 0 and positioner_id not in self._vpositioners:
                continue

            if positioner_id == 0:
                await asyncio.gather(
                    *[
                        vp.process_message(msg, positioner_id, command_id, uid)
                        for vp in self._vpositioners.values()
                    ]
                )
            else:
                vp = self._vpositioners[positioner_id]
                await vp.process_message(msg, positioner_id, command_id, uid)


class VirtualPositioner(StatusMixIn):
    """A virtual positioner that listen to CAN commands.

    An object of `.VirtualPositioner` represents a real positioner firmware.
    It knows it's own state at any time and replies.

    Parameters
    ----------

    positioner_id
        The ID of the positioner.
    centre
        A tuple of two floats indicating where on the focal plane the
        positioner is located.
    position
        A tuple of two floats indicating the angles of the alpha and beta arms.
    speed
        A tuple of two float indicating the RPM on the input for alpha and
        beta.
    firmware
        The firmware version, as a string with the format `AB.CD.EF`.

    """

    #: The initial status of the positioner. Represents a positioner that
    #: is not moving and is fully calibrated.
    _initial_status = (
        PositionerStatus.SYSTEM_INITIALIZED
        | PositionerStatus.DISPLACEMENT_COMPLETED
        | PositionerStatus.DISPLACEMENT_COMPLETED_ALPHA
        | PositionerStatus.DISPLACEMENT_COMPLETED_BETA
        | PositionerStatus.POSITION_RESTORED
        | PositionerStatus.DATUM_ALPHA_INITIALIZED
        | PositionerStatus.DATUM_BETA_INITIALIZED
        | PositionerStatus.MOTOR_ALPHA_CALIBRATED
        | PositionerStatus.MOTOR_BETA_CALIBRATED
        | PositionerStatus.CLOSED_LOOP_ALPHA
        | PositionerStatus.CLOSED_LOOP_BETA
    )

    def __init__(
        self,
        positioner_id: int,
        bus: Optional[VirtualBus] = None,
        centre: Optional[tuple] = None,
        position: Tuple[float, float] = (0.0, 0.0),
        speed: Optional[tuple] = None,
        firmware: str = "10.11.12",
    ):

        self.positioner_id = positioner_id
        self.centre = centre or (None, None)

        self.position = position
        self.speed = speed or (
            config["positioner"]["motor_speed"],
            config["positioner"]["motor_speed"],
        )

        self.firmware = firmware
        self._initial_firmware = firmware

        # To be used for a firmware upgrade.
        self._crc32 = 0
        self._firmware_size = 0
        self._firmware_received = b""

        self.number_trajectories = 1

        self.bus = bus

        StatusMixIn.__init__(
            self,
            PositionerStatus,
            initial_status=self._initial_status,
        )

    async def process_message(self, msg, positioner_id, command_id, uid):
        """Processes incoming commands from the bus."""

        command_id = CommandID(command_id)
        command = command_id.get_command_class()

        if positioner_id == 0 and not command.broadcastable:
            self.reply(
                command_id,
                uid,
                response_code=ResponseCode.INVALID_BROADCAST_COMMAND,
            )
            return

        if command_id == CommandID.GET_ID:
            self.reply(command_id, uid)

        elif command_id == CommandID.GET_FIRMWARE_VERSION:
            data_firmware = command.encode(self.firmware)  # type: ignore
            self.reply(command_id, uid, data=data_firmware)

        elif command_id == CommandID.GET_STATUS:
            data_status = utils.int_to_bytes(self.status)
            self.reply(command_id, uid, data=data_status)

        elif command_id == CommandID.GET_NUMBER_TRAJECTORIES:
            data_status = utils.int_to_bytes(self.number_trajectories)
            self.reply(command_id, uid, data=data_status)

        elif command_id in [
            CommandID.GO_TO_ABSOLUTE_POSITION,
            CommandID.GO_TO_RELATIVE_POSITION,
        ]:
            asyncio.create_task(self.process_goto(msg))

        elif command_id == CommandID.GET_ACTUAL_POSITION:
            data_position = command.encode(*self.position)  # type: ignore
            self.reply(command_id, uid, data=data_position)

        elif command_id == CommandID.SET_SPEED:
            data_speed = command.encode(*self.speed)  # type: ignore
            self.reply(command_id, uid, data=data_speed)

        elif command_id == CommandID.START_FIRMWARE_UPGRADE:
            if not self.is_bootloader():
                self.reply(
                    command_id,
                    uid,
                    response_code=ResponseCode.INVALID_COMMAND,
                )
                return

            try:
                data = msg.data
                firmware_size = utils.bytes_to_int(data[0:4], "u4")
                crc32 = utils.bytes_to_int(data[4:9], "u4")
            except Exception:
                self.reply(
                    command_id,
                    uid,
                    response_code=ResponseCode.INVALID_COMMAND,
                )
                return

            self._firmware_size = firmware_size
            self._crc32 = crc32
            self._firmware_received = b""

            self.reply(command_id, uid)

        elif command_id == CommandID.SEND_FIRMWARE_DATA:
            asyncio.create_task(self.process_firmware_data(uid, msg.data))

        else:
            # Should be a valid command or CommandID(command_id) would
            # have failed. Just return OK.
            self.reply(command_id, uid)

    def reply(self, command_id, uid, response_code=None, data=None):

        response_code = response_code or ResponseCode.COMMAND_ACCEPTED

        if isinstance(data, (bytearray, bytes)):
            data = [data]
        elif not data:
            data = [None]

        reply_id = utils.get_identifier(
            self.positioner_id,
            command_id,
            uid=uid,
            response_code=response_code,
        )

        for data_chunk in data:
            message = Message(
                arbitration_id=reply_id,
                is_extended_id=True,
                data=data_chunk,
            )
            # if self.notifier:
            #     self.notifier.bus.send(message)
            assert self.bus
            self.bus.send(message)

    async def process_firmware_data(self, uid, data):
        """Processes ``SEND_FIRMWARE_DATA`` commands."""

        command_id = CommandID.SEND_FIRMWARE_DATA

        if len(data) > 8:
            self.reply(
                command_id,
                uid,
                response_code=ResponseCode.VALUE_OUT_OF_RANGE,
            )
            return

        self._firmware_received += data

        fw_size = len(self._firmware_received)

        if fw_size > self._firmware_size:
            self.reply(
                command_id,
                uid,
                response_code=ResponseCode.VALUE_OUT_OF_RANGE,
            )
        elif fw_size == self._firmware_size:
            if not zlib.crc32(self._firmware_received) == self._crc32:
                self.reply(
                    command_id,
                    uid,
                    response_code=ResponseCode.VALUE_OUT_OF_RANGE,
                )
            else:
                self.firmware = self._firmware_received.decode("utf-8")[-8:]
                self.reply(command_id, uid)
        else:
            self.reply(command_id, uid)

    async def process_goto(self, message):
        """Process an absolute or relative goto command."""

        __, command_id, uid, __ = utils.parse_identifier(message.arbitration_id)
        command_id = CommandID(command_id)
        command = command_id.get_command_class()

        data = message.data
        alpha_move, beta_move = command.decode(data)  # type: ignore

        if command_id == CommandID.GO_TO_RELATIVE_POSITION:
            alpha_move += self.position[0]
            beta_move += self.position[0]

        target_alpha = self.position[0] + alpha_move
        target_beta = self.position[1] + beta_move

        if (
            target_alpha < 0
            or target_beta < 0
            or target_alpha > 360
            or target_beta > 360
        ):
            self.reply(command_id, uid, ResponseCode.VALUE_OUT_OF_RANGE)
            return

        if alpha_move == 0.0:
            alpha_move_time = 0.0
        else:
            alpha_move_time = int(
                utils.get_goto_move_time(alpha_move, self.speed[0]) / TIME_STEP
            )

        if beta_move == 0.0:
            beta_move_time = 0.0
        else:
            beta_move_time = int(
                utils.get_goto_move_time(beta_move, self.speed[1]) / TIME_STEP
            )

        self.reply(
            command_id,
            uid,
            ResponseCode.COMMAND_ACCEPTED,
            data=[
                utils.int_to_bytes(alpha_move_time, "i4")
                + utils.int_to_bytes(beta_move_time, "i4")
            ],
        )

        self.status ^= (
            PositionerStatus.DISPLACEMENT_COMPLETED
            | PositionerStatus.DISPLACEMENT_COMPLETED_ALPHA
            | PositionerStatus.DISPLACEMENT_COMPLETED_BETA
        )
        self.status |= (
            PositionerStatus.TRAJECTORY_ALPHA_RECEIVED
            | PositionerStatus.TRAJECTORY_BETA_RECEIVED
        )

        await asyncio.sleep(max(alpha_move * TIME_STEP, beta_move_time * TIME_STEP))

        self.position = (target_alpha, target_beta)

        self.status |= (
            PositionerStatus.DISPLACEMENT_COMPLETED
            | PositionerStatus.DISPLACEMENT_COMPLETED_ALPHA
            | PositionerStatus.DISPLACEMENT_COMPLETED_BETA
        )

    def reset(self):
        """Resets the positioner."""

        self.position = (0.0, 0.0)
        self.status = self._initial_status
        self.firmware = self._initial_firmware

    def is_bootloader(self):
        """Returns `True` if the positioner is in bootloader mode."""

        return self.firmware.split(".")[1] == "80"

    def set_bootloader(self, bootloader=True):
        """Sets the positioner in bootloader mode."""

        firmware_chunks = self.firmware.split(".")

        if bootloader:
            firmware_chunks[1] = "80"
            self.flags = BootloaderStatus
            self.status = BootloaderStatus.BOOTLOADER_INIT
        else:
            firmware_chunks[1] = self._initial_firmware.split(".")[1]
            self.flags = PositionerStatus
            self.status = self._initial_status

        self.firmware = ".".join(firmware_chunks)



#!/usr/bin/env python
# flake8: noqa
# type: ignore

import asyncio
import time

import click
import numpy
import pandas as pd

from kaiju import utils
from kaiju.robotGrid import RobotGridCalib
from sdsstools.daemonizer import cli_coro


# hardcoded defaults
MET_LED = 6  # leds 1 and 2
AP_LED = 16  # led 3
BOSS_LED = 8  # led 4
ANG_STEP = 0.1  # step size in degrees for path generation
EPS = ANG_STEP * 2  # absolute distance used for path smoothing
USE_SYNC_LINE = False  # whether or not to use the sync line for paths
GREED = 0.9  # parameter for MDP
PHOBIA = 0.1  # parameter for MDP
SMOOTH_PTS = 5  # number of points for smoothing paths
COLLISION_SHRINK = 0.08  # mm to shrink buffers by for path smoothing/simplification
PATH_DELAY = 1  # seconds of time in the future to send the first point
SAFE_BETA = [165, 195]
# MAX_ALPHA = 358  # set here for robot 444 without full range alpha travel
BAD_ROBOTS = [54, 463]  # put offline robots in this list?
DOWNSAMPLE = 100


def getTargetCoords(rg):
    # return the desired xyWok positions for the metrology
    # fiber for each robot, move this stuff to robot grid...
    positioner_id = []
    fibre_type = []
    hole_id = []
    xwok = []
    ywok = []
    alpha = []
    beta = []
    offline = []

    for r in rg.robotDict.values():

        # append metrology fiber info
        positioner_id.append(r.id)
        hole_id.append(r.holeID)
        fibre_type.append("Metrology")
        xwok.append(r.metWokXYZ[0])
        ywok.append(r.metWokXYZ[1])
        alpha.append(r.alpha)
        beta.append(r.beta)
        offline.append(int(r.isOffline))

        # append boss fiber info
        positioner_id.append(r.id)
        hole_id.append(r.holeID)
        fibre_type.append("BOSS")
        xwok.append(r.bossWokXYZ[0])
        ywok.append(r.bossWokXYZ[1])
        alpha.append(r.alpha)
        beta.append(r.beta)
        offline.append(int(r.isOffline))

        # append apogee fiber info
        positioner_id.append(r.id)
        hole_id.append(r.holeID)
        fibre_type.append("APOGEE")
        xwok.append(r.apWokXYZ[0])
        ywok.append(r.apWokXYZ[1])
        alpha.append(r.alpha)
        beta.append(r.beta)
        offline.append(int(r.isOffline))

    return pd.DataFrame(
        {
            "positioner_id": positioner_id,
            "hole_id": hole_id,
            "fibre_type": fibre_type,
            "xwok": xwok,
            "ywok": ywok,
            "alpha": alpha,
            "beta": beta,
            "offline": offline,
            "assigned": [1] * len(positioner_id),
        }
    )


def getRandomGrid(seed, danger=False, collisionBuffer=None, lefthand=False):
    numpy.random.seed(seed)
    if lefthand:
        alphaDestination = 350
        betaDestination = 190
    else:
        alphaDestination = 10
        betaDestination = 170

    rg = RobotGridCalib(ANG_STEP, EPS, seed)

    if collisionBuffer is not None:
        rg.setCollisionBuffer(collisionBuffer)

    # hardcode offline robots here? eventually get them from jaeger config?
    # or RobotGridCalib will load from positionerTable
    # example:

    rg.robotDict[54].setAlphaBeta(10.013, 169.9634)
    rg.robotDict[54].setDestinationAlphaBeta(10.013, 169.9634)
    rg.robotDict[54].isOffline = True

    rg.robotDict[463].setAlphaBeta(10.0103, 170.0181)
    rg.robotDict[463].setDestinationAlphaBeta(10.0103, 170.0181)
    rg.robotDict[463].isOffline = True

    # < no offline robots yet > #

    for robot in rg.robotDict.values():
        if robot.isOffline:
            continue

        if danger:
            # full range of motion
            if lefthand:
                robot.lefthanded = True
            # for now special handling for robot 444
            # note, i'm not checking after
            # decolliding grid, which maybe I should?
            else:
                robot.setXYUniform()
        else:
            alpha = numpy.random.uniform(0, 359.999)  # MAX_ALPHA)
            beta = numpy.random.uniform(SAFE_BETA[0], SAFE_BETA[1])
            if robot.id == 999:
                print("robot 999 alpha", alpha)
            robot.setAlphaBeta(alpha, beta)

        robot.setDestinationAlphaBeta(alphaDestination, betaDestination)

    if danger:
        rg.decollideGrid()
    else:
        if rg.getNCollisions() != 0:
            raise RuntimeError("supposedly safe grid is apparently collided!!!")

    return rg


async def setKaijuCurrent(fps, rg):
    """Get a kaiju robot grid initialized to the jaeger reported robot positions"""
    await fps.update_position()
    posArray = fps.get_positions()
    for rID, alpha, beta in posArray:
        robot = rg.robotDict[int(rID)]
        if robot.isOffline:
            continue
        robot.setAlphaBeta(alpha, beta)


async def ledOn(fps, devName, ledpower):
    on_value = 32 * int(1023 * (ledpower) / 100)
    device = fps.ieb.get_device(devName)
    await device.write(on_value)


async def ledOff(fps, devName):
    device = fps.ieb.get_device(devName)
    await device.write(0)


async def exposeFVC(fvc, exptime, fibre_data, nexp):
    from jaeger.exceptions import FVCError

    try:
        print("exposing FVC stack of %i" % (nexp))
        rawfname = await fvc.expose(exposure_time=exptime, stack=nexp)
        print("exposure complete: %s" % rawfname)
        fvc.process_fvc_image(rawfname, fibre_data, plot=True)
        print("image processing complete")
        positions = await fvc.fps.update_position()
        fvc.calculate_offsets(positions)
        print("calculcate offsets complete")
        await fvc.write_proc_image()
        print("image write complete")
    except FVCError as e:
        print("exposure failed with FVCError, continuing")


async def unwind(fps, speed, collisionBuffer):
    _seed = 0  # unimportant for unwind
    rg = getRandomGrid(seed=_seed, collisionBuffer=collisionBuffer)

    # set the robot grid to the current jaeger positions
    await setKaijuCurrent(fps, rg)
    if rg.getNCollisions() > 0:
        print("refuse to unwind, grid is kaiju collided!")
        return False

    # generate the path to fold
    tstart = time.time()
    rg.pathGenGreedy(stopIfDeadlock=False)
    print("unwind path generation took %.1f seconds" % (time.time() - tstart))

    # verify that the path generation was successful if not exit
    if rg.didFail:
        print("failed to unwind grid. deadlock in path generation")
        return False

    # smooth the paths
    toDestination, fromDestination = rg.getPathPair(
        speed=speed,
        smoothPoints=SMOOTH_PTS,
        collisionShrink=COLLISION_SHRINK,
        pathDelay=PATH_DELAY,
    )

    # check for collisions in path smoothing
    if rg.smoothCollisions:
        print("failed to unwind grid. collision in path smoothing")
        return False

    # command the path from the initial state (set by jaeger)
    # to the destination state (folded)
    tstart = time.time()
    print("sending unwind trajectory")
    await fps.send_trajectory(toDestination, use_sync_line=USE_SYNC_LINE)
    print("unwind finished, took %.1f seconds" % (time.time() - tstart))
    return True


@click.command()
@click.option(
    "--niter",
    default=1,
    type=click.IntRange(min=1),
    show_default=True,
    help="number of move iterations to perform",
)
@click.option(
    "--nexp",
    default=1,
    type=click.IntRange(min=0),
    show_default=True,
    help="number of repeated exposures for each move and backlit fiber combo",
)
@click.option(
    "--exptime",
    default=5,
    type=click.FloatRange(min=0),
    show_default=True,
    help="fvc exposure time, if 0, no exposures are taken",
)
@click.option("--seed", type=click.IntRange(min=0), help="random seed to use")
@click.option(
    "--speed",
    default=2,
    type=click.FloatRange(0.1, 2.9),
    show_default=True,
    help="speed of robot in RPM at output",
)
@click.option("--lh", is_flag=True, help="if passed, use lefthand robot kinematics")
@click.option(
    "--cb",
    default=2.1,
    type=click.FloatRange(1.5, 2.5),
    show_default=True,
    help="set collision buffer for all robots in grid",
)
@click.option(
    "--danger",
    is_flag=True,
    show_default=True,
    help="if passed, use full workspace for each robot, else limit to safe moves only",
)
@click.option(
    "--met",
    is_flag=True,
    show_default=True,
    help="if passed, get image of back illuminated metrology fibers",
)
@click.option(
    "--boss",
    is_flag=True,
    show_default=True,
    help="if passed, get image of back illuminated boss fibers",
)
@click.option(
    "--apogee",
    is_flag=True,
    show_default=True,
    help="if passed, get image of back illuminated apogee fibers",
)
@click.option(
    "--allfibers",
    is_flag=True,
    show_default=True,
    help="if passed, get image of all fibers back illuminated simultaneously",
)
@click.option(
    "--mdp",
    is_flag=True,
    show_default=True,
    help="if passed, use markov decision process "
    "for path generation, otherwise a greedy algorithm is used",
)
@click.option(
    "--nomove",
    is_flag=True,
    show_default=True,
    help="if passed, do not execute a move, leave array as it is",
)
@click.option(
    "--simpath",
    is_flag=True,
    show_default=True,
    help="if passed, simulate the trajectory but don't send it.  make a movie",
)
@cli_coro()
async def robotcalib(
    niter,
    exptime,
    speed,
    cb,
    nexp,
    danger=False,
    lh=True,
    seed=None,
    met=False,
    boss=False,
    apogee=False,
    allfibers=False,
    mdp=False,
    nomove=False,
    simpath=False,
):
    if seed is None:
        seed = numpy.random.randint(0, 30000)

    if not simpath:
        # unwind if we aren't simulating!!!
        from jaeger import config, log
        from jaeger.exceptions import FVCError, TrajectoryError
        from jaeger.fvc import FVC

        log.sh.setLevel(20)

        fvc = FVC(config["observatory"])
        fps = fvc.fps
        await fps.initialise()
        await fps.unlock()

        ######## UNWIND GRID #############

        success = await unwind(fps, speed, cb - 0.1)
        if not success:
            return  # unwind failed
        ########### UNWIND FINISHED ##############

    ########### BEGIN CALIBRATION LOOP #######
    movesExecuted = 0
    nErrorsForward = 0
    badRobotForward = []
    nErrorsReverse = 0
    badRobotReverse = []
    for ii in range(niter):
        seed += 1
        print(
            "\n----------------------\nITER %i of %i (seed=%i)\n----------------------"
            % (ii + 1, niter, seed)
        )

        # begin searching for a valid path
        # this is easy when things are safe
        # and easier for smaller collision buffers using the MDP algorithm
        replacedRobotList = []
        rg = getRandomGrid(seed=seed, danger=danger, collisionBuffer=cb, lefthand=lh)

        # try 5 times to find valid path
        # before giving up and moving on
        for jj in range(5):
            if nomove:
                # we're not moving!
                await setKaijuCurrent(fps, rg)
                targetCoords = getTargetCoords(rg)
                break
            print("path gen attempt %i" % jj)

            # get the expected coords for each fiber on each robot
            targetCoords = getTargetCoords(rg)

            tstart = time.time()
            if mdp:
                rg.pathGenMDP(GREED, PHOBIA)
            else:
                rg.pathGenGreedy(stopIfDeadlock=False)
            print(
                "attempt %i path generation took %.1f seconds"
                % (jj, (time.time() - tstart))
            )

            if not rg.didFail:
                break

            dlrobots = rg.deadlockedRobots()
            print("%i deadlocked robots" % len(dlrobots))
            if len(dlrobots) > 6:
                print("too many deadlocks to resolve")
                break

            replaceableRobots = list(set(dlrobots) - set(BAD_ROBOTS))
            nextReplacement = numpy.random.choice(replaceableRobots)
            replacedRobotList.append(nextReplacement)
            rg = getRandomGrid(
                seed=seed, danger=danger, collisionBuffer=cb, lefthand=lh
            )
            for robotID in replacedRobotList:
                robot = rg.robotDict[robotID]
                robot.setXYUniform()
            rg.decollideGrid()

        if not nomove:
            if rg.didFail:
                print("failed to generate valid paths, skipping to next iteration")
                continue

            print("valid paths found")
            toDestination, fromDestination = rg.getPathPair(
                speed=speed,
                smoothPoints=SMOOTH_PTS,
                collisionShrink=COLLISION_SHRINK,
                pathDelay=PATH_DELAY,
            )

            # print max steps in paths
            maxPoints = -1
            for alphaBetaPath in toDestination.values():
                for path in alphaBetaPath.values():
                    if len(path) > maxPoints:
                        maxPoints = len(path)
            print("maximum path points %i" % maxPoints)

            if simpath:
                ## simulate only, don't proceed further
                movesExecuted += 1
                interpSteps = len(
                    list(rg.robotDict.values())[0].interpSimplifiedAlphaPath
                )
                print("interpsteps", interpSteps)
                print(
                    "simulating path with %.5e smooth collisions per step"
                    % (rg.smoothCollisions / interpSteps)
                )
                print("smooth collided robots", rg.smoothCollidedRobots)
                for rid in rg.smoothCollidedRobots:
                    r = rg.robotDict[rid]
                    utils.plotTraj(r, "simpath_%i" % ii, dpi=250)
                # utils.plotPaths(rg, downsample=DOWNSAMPLE, filename="simpath_%i.mp4"%ii)
                continue

            if rg.smoothCollisions:
                print(
                    "%i smooth collisions, skipping to next iteration"
                    % (rg.smoothCollisions)
                )
                continue

            movesExecuted += 1
            tstart = time.time()
            print("sending path fold-->target")
            try:
                await fps.send_trajectory(fromDestination, use_sync_line=USE_SYNC_LINE)
                print("move complete, duration %.1f" % (time.time() - tstart))
            except TrajectoryError as e:
                nErrorsForward += 1
                badRobotForward.append(e.trajectory.failed_positioners)
                print("TRAJECTORY ERROR moving fold-->target")
                print("failed positioners: ", str(e.trajectory.failed_positioners))
                return

                # print("attempting to recover from trajectory error with unwind")
                # # shrink collision buffer
                # _cbShrink = cb - 0.1
                # await fps.unlock()
                # # success = await unwind(fps, speed, _cbShrink)
                # success = False
                # if success:
                #     print("unwind worked skipping to next iteration")
                #     continue
                # else:
                #     print("moves executed", movesExecuted)
                #     print("number of traj errors fold-->target", nErrorsForward)
                #     print("err robots fold-->target", badRobotForward)
                #     print("number of traj errors target-->fold", nErrorsReverse)
                #     print("err robots target-->fold", badRobotReverse)
                #     return  # exit routine

        else:
            print("not moving array --nomove flag passed")

        ### turn all LEDs off ###
        await ledOff(fps, "led1")
        await ledOff(fps, "led2")
        await ledOff(fps, "led3")
        await ledOff(fps, "led4")
        await asyncio.sleep(1)

        if exptime > 0 and nexp != 0:
            if True not in [met, apogee, boss, allfibers]:
                # no fibers illuminated, expose anyway
                await exposeFVC(fvc, exptime, targetCoords)

            else:
                # take a single exposure for each fiber wanted
                if met:
                    print("back illuminating metrology")
                    await ledOn(fps, "led1", MET_LED)
                    await ledOn(fps, "led2", MET_LED)
                    await asyncio.sleep(1)
                    await exposeFVC(fvc, exptime, targetCoords, nexp)
                    await ledOff(fps, "led1")
                    await ledOff(fps, "led2")
                    await asyncio.sleep(1)
                if boss:
                    print("back illuminating boss")
                    await ledOn(fps, "led4", BOSS_LED)
                    await asyncio.sleep(1)
                    await exposeFVC(fvc, exptime, targetCoords, nexp)
                    await ledOff(fps, "led4")
                    await asyncio.sleep(1)
                if apogee:
                    print("back illuminating apogee")
                    await ledOn(fps, "led3", AP_LED)
                    await asyncio.sleep(1)
                    await exposeFVC(fvc, exptime, targetCoords, nexp)
                    await ledOff(fps, "led3")
                    await asyncio.sleep(1)
                if allfibers:
                    print("back illuminating all fibers")
                    await ledOn(fps, "led1", MET_LED)
                    await ledOn(fps, "led2", MET_LED)
                    await ledOn(fps, "led4", BOSS_LED)
                    await ledOn(fps, "led3", AP_LED)
                    await asyncio.sleep(1)
                    await exposeFVC(fvc, exptime, targetCoords, nexp)
                    await ledOff(fps, "led1")
                    await ledOff(fps, "led2")
                    await ledOff(fps, "led3")
                    await ledOff(fps, "led4")
                    await asyncio.sleep(1)

        ### send path back to lattice ####
        if not nomove:
            tstart = time.time()
            print("sending path target-->fold")
            try:
                await fps.send_trajectory(toDestination, use_sync_line=USE_SYNC_LINE)
                print("move complete, duration %.1f" % (time.time() - tstart))
            except TrajectoryError as e:
                nErrorsReverse += 1
                badRobotReverse.append(e.trajectory.failed_positioners)
                print("TRAJECTORY ERROR moving target-->fold")
                print("failed positioners: ", str(e.trajectory.failed_positioners))
                return

                # print("attempting to recover from trajectory error with unwind")
                # # shrink collision buffer
                # _cbShrink = cb - 0.1
                # await fps.unlock()
                # # success = await unwind(fps, speed, _cbShrink)
                # success = False
                # if success:
                #     # unwind worked move to next iteration
                #     print("unwind worked skipping to next iteration")
                #     continue
                # else:
                #     print("moves executed", movesExecuted)
                #     print("number of traj errors fold-->target", nErrorsForward)
                #     print("err robots fold-->target", badRobotForward)
                #     print("number of traj errors target-->fold", nErrorsReverse)
                #     print("err robots target-->fold", badRobotReverse)
                #     return  # exit routine

    print("\n-------\nend script\n-------\n")
    print("moves executed", movesExecuted)
    print("number of traj errors fold-->target", nErrorsForward)
    print("err robots fold-->target", badRobotForward)
    print("number of traj errors target-->fold", nErrorsReverse)
    print("err robots target-->fold", badRobotReverse)


if __name__ == "__main__":
    asyncio.run(robotcalib())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-04
# @Filename: power.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING, Any

import click

from clu.parsers.click import pass_args
from drift import Relay

from jaeger import config
from jaeger.ieb import IEB, _get_category_data
from jaeger.testing import VirtualFPS

from . import jaeger_parser


if TYPE_CHECKING:
    from clu.command import Command

    from jaeger import FPS

    from .. import JaegerActor


__all__ = ["power"]


async def switch(
    command: Command[JaegerActor],
    fps: FPS,
    devices: list[str],
    on: bool | None = None,
    cycle: bool = False,
    delay: float = 1.0,
    force: bool = False,
):
    """Switches the status of an on/off device."""

    ieb = fps.ieb
    assert isinstance(ieb, IEB)

    if len(devices) == 0:
        return command.fail(error="No devices specified.")

    for idev, device in enumerate(devices):

        if device.upper() in config["ieb"]["disabled_devices"] and on is True:
            if force is False:
                command.warning(text=f"{device} is disabled. Skipping.")
                continue
            else:
                command.warning(text=f"{device} is disabled but overriding.")

        if len(devices) > 1 and idev != 0:
            await asyncio.sleep(delay)

        if cycle:
            on = False

        try:
            device_obj = ieb.get_device(device)
            assert isinstance(device_obj, Relay)
            dev_name = device_obj.name
        except ValueError:
            return command.fail(error=f"Cannot find device {device!r}.")

        if device_obj.module.mode not in ["holding_register", "coil"]:
            return command.fail(error=f"{dev_name!r} is not an output.")

        if on is None:  # The --on/--off was not passed
            current_status = (await device_obj.read())[0]
            if current_status == "closed":
                on = False
            elif current_status == "open":
                on = True
            else:
                return command.fail(
                    error=f"Invalid status for device {dev_name!r}: {current_status!r}."
                )

        try:
            if on is True:
                await device_obj.close()
            elif on is False:
                await device_obj.open()
        except Exception:
            return command.fail(error=f"Failed to set status of device {dev_name!r}.")

        if cycle:
            command.write("d", text="Waiting 1 second before powering up.")
            await asyncio.sleep(1)
            try:
                await device_obj.close()
            except Exception:
                return command.fail(
                    error=f"Failed to power device {dev_name!r} back on."
                )

        # If we read the status immediately sometimes we still get the old one.
        # Sleep a bit to avoid that.
        await asyncio.sleep(0.2)

        status = "on" if (await device_obj.read())[0] == "closed" else "off"

        message: dict[str, Any] = {"text": f"Device {dev_name!r} is now {status!r}."}
        category = device_obj.category
        if category:
            message[category] = await _get_category_data(command, category)

        command.info(message=message)

    return command.finish()


async def _power_sequence(command, ieb, seq, mode="on", delay=3) -> bool:
    """Applies the power on/off sequence."""

    # To speed up tests
    if isinstance(command.actor.fps, VirtualFPS):
        delay = 0.01

    relay_result = "closed" if mode == "on" else "open"

    command.info(text=f"Running power {mode} sequence")

    # First check that the SYNC line is open.
    sync = ieb.get_device("SYNC")
    if (await sync.read())[0] == "closed":
        command.debug(text="SYNC line is high. Opening it.")
        await sync.open()

    if (await sync.read())[0] != "open":
        command.fail(error="Failed opening SYNC line.")
        return False

    command.debug(power_sync=[False])

    disabled = config["ieb"]["disabled_devices"]

    for devname in seq:

        do_delay = False
        if isinstance(devname, str):
            if devname.upper() in config["ieb"]["disabled_devices"]:
                command.warning(text=f"{devname} is disabled. Skipping.")
                continue

            dev = ieb.get_device(devname)
            category = dev.category.lower()

            if (await dev.read())[0] == relay_result:
                command.debug(text=f"{devname} already powered {mode}.")
            else:
                command.debug(text=f"Powering {mode} {devname}.")
                await dev.close() if mode == "on" else await dev.open()
                await asyncio.sleep(0.2)
                if (await dev.read())[0] != relay_result:
                    command.fail(error=f"Failed powering {mode} {devname}.")
                    return False

                do_delay = True

            command.debug({category: await _get_category_data(command, category)})

        elif isinstance(devname, (tuple, list)):
            devname = list([dn for dn in devname if dn not in disabled])
            devs = [ieb.get_device(dn) for dn in devname]
            category = devs[0].category.lower()

            status = list(await asyncio.gather(*[dev.read() for dev in devs]))

            keep = []
            for ii, res in enumerate(status):
                if res[0] == relay_result:
                    command.debug({"text": f"{devname[ii]} already powered {mode}."})
                    continue
                keep.append(ii)

            devs = [devs[ii] for ii in keep]
            devname = [devname[ii] for ii in keep]

            if len(devs) > 0:
                command.debug(text=f"Powering {mode} {', '.join(devname)}")
                await asyncio.gather(
                    *[dev.close() if mode == "on" else dev.open() for dev in devs]
                )
                do_delay = True

            status = list(await asyncio.gather(*[dev.read() for dev in devs]))
            for ii, res in enumerate(status):
                if res[0] != relay_result:
                    command.fail(error=f"Failed powering {mode} {devname[ii]}.")
                    return False

            command.debug({category: await _get_category_data(command, category)})

        else:
            command.fail(error=f"Invalid relay {devname!r}.")
            return False

        if do_delay:
            await asyncio.sleep(delay)

    return True


@jaeger_parser.group()
@pass_args()
def power(command, fps):
    """Runs the power on/off sequences and powers individual devices."""

    pass


@power.command()
@click.argument("DEVICES", type=str, nargs=-1)
@click.option("--gfas", is_flag=True, help="Power on the GFAs.")
@click.option(
    "--delay",
    type=float,
    default=3,
    help="When powering multiple devices, the delay to wait between them.",
)
@click.option(
    "--force",
    is_flag=True,
    help="Forces a device to turn on/off even if disabled.",
)
async def on(
    command: Command[JaegerActor],
    fps: FPS,
    devices: list[str],
    gfas: bool = False,
    delay: float = 3,
    force: bool = True,
):
    """Powers on all the FPS IEB components or a specific device."""

    if len(devices) > 0:
        return await switch(command, fps, devices, on=True, force=force, delay=delay)

    ieb = fps.ieb
    if not isinstance(ieb, IEB) or ieb.disabled:
        return command.fail(error="IEB is not conencted or is disabled.")

    # Sequence of relays to power on. Tuples indicate relays that can be powered
    # on concurrently.
    on_seq = [
        ("CM1", "CM2", "CM3", "CM4", "CM5", "CM6"),
        "PS1",
        "PS2",
        "PS3",
        "PS4",
        "PS5",
        "PS6",
        "NUC1",
        "NUC2",
        "NUC3",
        "NUC4",
        "NUC5",
        "NUC6",
    ]

    if gfas is True:
        on_seq += [
            "GFA1",
            "GFA2",
            "GFA3",
            "GFA4",
            "GFA5",
            "GFA6",
        ]

    if not (await _power_sequence(command, ieb, on_seq, mode="on", delay=int(delay))):
        return

    command.info("Waiting 15 seconds and reloading positioners.")
    await asyncio.sleep(15)

    await fps.initialise()

    return command.finish(text="Power on sequence complete.")


@power.command()
@click.argument("DEVICES", type=str, nargs=-1)
@click.option("--nucs", is_flag=True, help="Also power down NUCs")
@click.option(
    "--delay",
    type=float,
    default=1,
    help="When powering multiple devices, the delay to wait between them.",
)
async def off(
    command: Command[JaegerActor],
    fps: FPS,
    devices: list[str],
    nucs: bool = False,
    delay: float = 1,
):
    """Powers off all the FPS IEB components or a specific device."""

    if len(devices) > 0:
        return await switch(command, fps, devices, on=False, force=True, delay=delay)

    command.info(text="Turning pollers off.")
    await fps.pollers.stop()

    ieb = fps.ieb
    if not isinstance(ieb, IEB) or ieb.disabled:
        return command.fail(error="IEB is not conencted or is disabled.")

    # Sequence of relays to power off. Tuples indicate relays that can be powered
    # off concurrently.
    off_seq = [
        "PS1",
        "PS2",
        "PS3",
        "PS4",
        "PS5",
        "PS6",
        "GFA1",
        "GFA2",
        "GFA3",
        "GFA4",
        "GFA5",
        "GFA6",
    ]

    if nucs:
        off_seq += ["NUC1", "NUC2", "NUC3", "NUC4", "NUC5", "NUC6"]

    if not (await _power_sequence(command, ieb, off_seq, mode="off", delay=int(delay))):
        return

    command.info("Waiting 15 seconds and reloading positioners.")
    await asyncio.sleep(15)

    await fps.initialise()

    return command.finish(text="Power off sequence complete.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-28
# @Filename: kaiju.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import time
import warnings

from typing import TYPE_CHECKING, Optional, cast

import numpy

from jaeger import config, log
from jaeger.exceptions import JaegerError, JaegerUserWarning, TrajectoryError
from jaeger.utils.helpers import run_in_executor


if TYPE_CHECKING:
    from kaiju import RobotGridCalib

    from jaeger.fps import FPS


__all__ = [
    "get_robot_grid",
    "dump_robot_grid",
    "decollide",
    "get_path_pair",
    "get_snapshot",
    "unwind",
    "explode",
    "get_path_pair_in_executor",
    "decollide_in_executor",
    "check_trajectory",
]


def warn(message):
    warnings.warn(message, JaegerUserWarning)


def get_robot_grid(fps: FPS | None, seed: int | None = None, collision_buffer=None):
    """Returns a new robot grid with the destination set to the lattice position.

    If an initialised instance of the FPS is available, disabled robots will be
    set offline in the grid at their current positions.

    """

    from kaiju.robotGrid import RobotGridCalib

    if seed is None:
        t = 1000 * time.time()
        seed = int(int(t) % 2**32 / 1000)

    kaiju_config = config["kaiju"]
    ang_step = kaiju_config["ang_step"]
    collision_buffer = collision_buffer or kaiju_config["collision_buffer"]
    alpha0, beta0 = kaiju_config["lattice_position"]
    epsilon = ang_step * kaiju_config["epsilon_factor"]

    if collision_buffer < 1.5:
        raise JaegerError("Invalid collision buffer < 1.5.")

    log.debug(f"Creating RobotGridCalib with stepSize={ang_step}, epsilon={epsilon}.")

    robot_grid = RobotGridCalib(stepSize=ang_step, epsilon=epsilon, seed=seed)
    robot_grid.setCollisionBuffer(collision_buffer)

    # TODO: This is a bit hacky. Kaiju doesn't have a collisionBuffer anymore
    # as collision buffers are per robot, but I want to keep this information
    # for when I dump and reload robot grids.
    robot_grid.collisionBuffer = collision_buffer

    for robot in robot_grid.robotDict.values():
        if fps is not None and robot.id in fps and fps[robot.id].disabled is True:
            positioner = fps[robot.id]
            robot.setDestinationAlphaBeta(positioner.alpha, positioner.beta)
            robot.setAlphaBeta(positioner.alpha, positioner.beta)
            robot.isOffline = True
        else:
            robot.setDestinationAlphaBeta(alpha0, beta0)
            robot.setAlphaBeta(alpha0, beta0)

    return robot_grid


def dump_robot_grid(robot_grid: RobotGridCalib) -> dict:
    """Dump the information needed to restore a robot grid into a dictionary."""

    data = {}

    data["collision_buffer"] = robot_grid.collisionBuffer
    data["grid"] = {}

    for robot in robot_grid.robotDict.values():
        alpha = robot.alpha
        beta = robot.beta
        destinationAlpha = robot.destinationAlpha
        destinationBeta = robot.destinationBeta

        data["grid"][robot.id] = (
            alpha,
            beta,
            destinationAlpha,
            destinationBeta,
            robot.isOffline,
        )

    return data


def load_robot_grid(data: dict, set_destination: bool = True) -> RobotGridCalib:
    """Restores a robot grid from a dump."""

    collision_buffer = data["collision_buffer"]
    robot_grid = get_robot_grid(None, collision_buffer=collision_buffer)

    for robot in robot_grid.robotDict.values():
        data_robot = data["grid"][robot.id]
        robot.setAlphaBeta(data_robot[0], data_robot[1])
        if set_destination:
            robot.setDestinationAlphaBeta(data_robot[2], data_robot[3])
        if data_robot[4] is True:
            robot.isOffline = True

    return robot_grid


def decollide(
    robot_grid: Optional[RobotGridCalib] = None,
    data: Optional[dict] = None,
    simple: bool = False,
    decollide_grid_fallback: bool = False,
    priority_order: list[int] = [],
) -> tuple[RobotGridCalib | dict, list[int]]:
    """Decollides a potentially collided grid. Raises on fail.

    Parameters
    ----------
    robot_grid
        The Kaiju ``RobotGridCalib`` instance to decollide.
    data
        A dictionary of data that can be used to reload a Kaiju robot grid
        using `.load_robot_grid`. This is useful if the function is being
        run in an executor.
    simple
        Runs ``decollideGrid()`` and returns.
    decollide_grid_fallback
        If `True`, runs ``decollideGrid()`` if the positioner-by-positioner
        decollision fails.
    priority_list
        A sorted list of positioner IDs with the order of which positioners to
        try to keep at their current positions. Positioners earlier in the list
        will be decollided last. Ignore in case of ``simple=True``.

    Returns
    -------
    grid,decollided
        If ``robot_grid`` is passed, returns the same grid instance after decollision.
        If ``data`` is passed, returns a dictionary describing the decollided grid
        that can be used to recreate a grid using `.load_robot_grid`. Also returns
        the list of robots that have been decollided.

    """

    if robot_grid is not None and data is not None:
        raise JaegerError("robot_grid and data are mutually exclusive.")

    if data is not None:
        robot_grid = load_robot_grid(data)

    assert robot_grid is not None

    if simple:
        collided = robot_grid.getCollidedRobotList()
        robot_grid.decollideGrid()
        if len(robot_grid.getCollidedRobotList()) > 0:
            raise JaegerError("Failed decolliding grid.")

        if data is not None:
            return dump_robot_grid(robot_grid), collided
        else:
            return robot_grid, collided

    # First pass. If collided, decollide each robot one by one.
    collided = robot_grid.getCollidedRobotList()

    # Sort collided robots by priority order.
    collided = sorted(
        collided,
        key=lambda x: len(priority_order) - priority_order.index(x)
        if x in priority_order
        else -1,
    )

    decollided = []
    if len(collided) > 0:
        for robot_id in collided:
            if robot_grid.isCollided(robot_id):
                if robot_grid.robotDict[robot_id].isOffline:
                    continue
                robot_grid.decollideRobot(robot_id)
                decollided.append(robot_id)  # Even if we failed it may have moved.
                if robot_grid.isCollided(robot_id):
                    raise JaegerError(f"Failed decolliding positioner {robot_id}.")

    # Second pass. If still collided, try a grid decollision.
    if len(robot_grid.getCollidedRobotList()) > 0:
        if decollide_grid_fallback:
            warn("Grid is still colliding. Attempting full grid decollision.")
            robot_grid.decollideGrid()
            if robot_grid.getCollidedRobotList() is not False:
                raise JaegerError("Failed decolliding grid.")
            # We don't know which robots were decollided so assume all collided
            # robots have moved.
            decollided = collided
        else:
            raise JaegerError("Failed decolliding grid.")

    if data is not None:
        return dump_robot_grid(robot_grid), decollided
    else:
        return robot_grid, decollided


def get_path_pair(
    robot_grid: Optional[RobotGridCalib] = None,
    data: Optional[dict] = None,
    path_generation_mode: str | None = None,
    ignore_did_fail: bool = False,
    explode_deg: float = 5,
    explode_positioner_id: int | None = None,
    speed=None,
    smooth_points=None,
    path_delay=None,
    collision_shrink=None,
    greed: float | None = None,
    phobia: float | None = None,
    stop_if_deadlock: bool = False,
    ignore_initial_collisions: bool = False,
) -> tuple:
    """Runs path generation and returns the to and from destination paths.

    Parameters
    ----------
    robot_grid
        The Kaiju ``RobotGridCalib`` instance to decollide.
    data
        A dictionary of data that can be used to reload a Kaiju robot grid
        using `.load_robot_grid`. This is useful if the function is being
        run in an executor.
    path_generation_mode
        Defines the path generation algorithm to use.
        Either ``greedy``, ``mdp``, ``explode`` or ``explode_one``. If
        `None`, defaults to ``kaiju.default_path_generator``.
    ignore_did_fail
        Generate paths even if path generation failed (i.e., deadlocks).
    explode_deg
        Degrees for ``pathGenExplode``.
    explode_positioner_id
        The positioner to explode.
    speed, smooth_points, path_delay, collision_shrink
        Kaiju parameters to pass to ``getPathPair``. Otherwise uses the default
        configuration values.
    phobia, greed
        Parameters for ``pathGenMDP``. If not set uses ``kaiju`` configuration
        values.
    stop_if_deadlock
        If `True`, detects deadlocks early in the path and returns shorter
        trajectories (at the risk of some false positive deadlocks).
    ignore_initial_collisions
        If `True`, does not fail if the initial state is collided. To be used
        only for offsets.

    Returns
    -------
    paths
        A tuple with the to destination path, from destination path, whether path
        generation failed, and the list of deadlocks

    """

    if path_generation_mode is None:
        path_generation_mode = cast(str, config["kaiju"]["default_path_generator"])

    if robot_grid is not None and data is not None:
        raise JaegerError("robot_grid and data are mutually exclusive.")

    if data is not None:
        set_destination = False if path_generation_mode == "explode" else True
        robot_grid = load_robot_grid(data, set_destination=set_destination)

    assert robot_grid is not None

    deadlocks = []
    if path_generation_mode == "explode":
        log.debug(f"Running pathGenExplode with explode_deg={explode_deg}.")
        robot_grid.pathGenExplode(explode_deg)

    elif path_generation_mode == "explode_one":
        log.debug(
            f"Running pathGenExplodeOne with explode_deg={explode_deg}, "
            f"explode_positioner_id={explode_positioner_id}."
        )
        robot_grid.pathGenExplodeOne(explode_deg, explode_positioner_id)

    elif path_generation_mode == "greedy":
        log.debug(f"Running pathGenGreedy with stopIfDeadlock={stop_if_deadlock}.")
        robot_grid.pathGenGreedy(
            stopIfDeadlock=stop_if_deadlock,
            ignoreInitialCollisions=ignore_initial_collisions,
        )

    elif path_generation_mode == "mdp":
        greed = greed or config["kaiju"]["greed"]
        phobia = phobia or config["kaiju"]["phobia"]
        log.debug(f"Running pathGenMDP with phobia={phobia}, greed={greed}.")
        robot_grid.pathGenMDP2(
            greed=greed,
            phobia=phobia,
            ignoreInitialCollisions=ignore_initial_collisions,
        )

    else:
        raise ValueError(f"Invalid path_generation_mode={path_generation_mode!r}.")

    if path_generation_mode in ["greedy", "mdp"]:
        # Check for deadlocks.
        deadlocks = robot_grid.deadlockedRobots()
        if robot_grid.didFail and ignore_did_fail is False:
            return (None, None, robot_grid.didFail, deadlocks)

    speed = speed or config["kaiju"]["speed"]
    smooth_points = smooth_points or config["kaiju"]["smooth_points"]
    collision_shrink = collision_shrink or config["kaiju"]["collision_shrink"]
    path_delay = path_delay or config["kaiju"]["path_delay"]

    log.debug(
        f"Running getPathPair with speed={speed}, smoothPoints={smooth_points}, "
        f"collisionShrink={collision_shrink}, pathDelay={path_delay}."
    )

    to_destination, from_destination = robot_grid.getPathPair(
        speed=speed,
        smoothPoints=smooth_points,
        collisionShrink=collision_shrink,
        pathDelay=path_delay,
    )

    return (
        to_destination,
        from_destination,
        robot_grid.didFail,
        deadlocks,
    )


async def get_path_pair_in_executor(robot_grid: RobotGridCalib, **kwargs):
    """Calls `.get_path_pair` with a process executor."""

    data = dump_robot_grid(robot_grid)
    traj_data = await run_in_executor(
        get_path_pair,
        data=data,
        executor="process",
        **kwargs,
    )

    return traj_data


async def decollide_in_executor(
    robot_grid: RobotGridCalib, **kwargs
) -> tuple[RobotGridCalib, list[int]]:
    """Calls `.decollide` with a process executor."""

    data = dump_robot_grid(robot_grid)
    decollided_data, collided = await run_in_executor(
        decollide,
        data=data,
        executor="process",
        **kwargs,
    )

    return load_robot_grid(decollided_data), collided


async def unwind(
    current_positions: dict[int, tuple[float | None, float | None]],
    collision_buffer: float | None = None,
    disabled: list[int] = [],
    force: bool = False,
):
    """Folds all the robots to the lattice position.

    This coroutine uses a process pool executor to run Kaiju routines.

    """

    alpha0, beta0 = config["kaiju"]["lattice_position"]

    # We create the data directly since it's simple. This should be a bit faster
    # than creating a grid and dumping it.
    data = {"collision_buffer": collision_buffer, "grid": {}}
    for pid, (alpha, beta) in current_positions.items():
        data["grid"][int(pid)] = (alpha, beta, alpha0, beta0, pid in disabled)

    (to_destination, _, did_fail, deadlocks) = await run_in_executor(
        get_path_pair,
        data=data,
        path_generation_mode="greedy" if force is True else None,
        ignore_did_fail=force,
        stop_if_deadlock=force,
        executor="process",
    )
    if did_fail:
        if force is False:
            raise TrajectoryError(
                "Failed generating a valid unwind trajectory. "
                f"{len(deadlocks)} deadlocks were found ({deadlocks})."
            )
        else:
            log.warning("Deadlocks found in unwind but proceeding anyway.")

    return to_destination


async def explode(
    current_positions: dict[int, tuple[float | None, float | None]],
    explode_deg=20.0,
    collision_buffer: float | None = None,
    disabled: list[int] = [],
    positioner_id: int | None = None,
):
    """Explodes the grid by a number of degrees.

    This coroutine uses a process pool executor to run Kaiju routines.

    """

    alpha0, beta0 = config["kaiju"]["lattice_position"]

    data = {"collision_buffer": collision_buffer, "grid": {}}
    for pid, (alpha, beta) in current_positions.items():
        data["grid"][int(pid)] = (alpha, beta, alpha0, beta0, pid in disabled)

    if positioner_id is not None:
        path_generation_mode = "explode_one"
    else:
        path_generation_mode = "explode"

    (to_destination, *_) = await run_in_executor(
        get_path_pair,
        data=data,
        path_generation_mode=path_generation_mode,
        explode_deg=explode_deg,
        explode_positioner_id=positioner_id,
        ignore_did_fail=False,
        executor="process",
    )

    return to_destination


def get_snapshot_async(
    path: str,
    robot_grid: Optional[RobotGridCalib] = None,
    data: Optional[dict] = None,
    highlight: int | None = None,
    title: str | None = None,
):
    """Creates an FPS snapshot and saves it to disk. To be used with an executor.

    Parameters
    ----------
    path
        The path where to save the file.
    robot_grid
        The Kaiju ``RobotGridCalib`` instance to plot.
    data
        A dictionary of data that can be used to reload a Kaiju robot grid
        using `.load_robot_grid`. This is useful if the function is being
        run in an executor.
    highlight
        Robot to highlight.
    title
        A title for the plot.

    """

    import matplotlib.pyplot as plt

    if robot_grid is not None and data is not None:
        raise JaegerError("robot_grid and data are mutually exclusive.")

    if data is not None:
        robot_grid = load_robot_grid(data)

    assert robot_grid is not None

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".+array interface is deprecated.+")
        ax = robot_grid.plot_state(highlightRobot=highlight)

    assert ax is not None

    if title is not None:
        ax.set_title(title)

    plt.tight_layout()

    ax.figure.savefig(path)

    plt.close("all")


async def get_snapshot(
    path: str,
    fps: FPS | None = None,
    collision_buffer: float | None = None,
    highlight: int | None = None,
    title: str | None = None,
):
    """Plots a snapshot of the FPS and saves it to disk."""

    from jaeger.fps import FPS

    fps = fps or FPS.get_instance()
    if fps.initialised is False:
        await fps.initialise()

    await fps.update_position()

    if len(fps.positioners) == 0:
        raise ValueError("No positioners connected.")

    data = {"collision_buffer": collision_buffer, "grid": {}}
    for pid in fps.positioners.keys():
        data["grid"][int(pid)] = (
            fps[pid].alpha,
            fps[pid].beta,
            0,
            0,
            fps[pid].disabled,
        )

    await run_in_executor(
        get_snapshot_async,
        path,
        data=data,
        highlight=highlight,
        title=title,
        executor="process",
    )

    return True


async def check_trajectory(
    trajectory: dict,
    current_positions: dict | None = None,
    fps: FPS | None = None,
    atol=1.0,
) -> bool:
    """Checks that the current position matches the starting point of a trajectory.

    Parameters
    ----------
    trajectory
        The dictionary with the trajectory to check.
    current_positions
        A mapping of positioner ID to ``(alpha, beta)`` with the current arrangement
        of the FPS array.
    fps
        If ``current_positions`` is not passed, the `.FPS` instance is used to
        determine the current arrangement.

    """

    if current_positions is None:
        if fps:
            await fps.update_position()
            current_positions = fps.get_positions_dict(ignore_disabled=True)
        else:
            raise RuntimeError("Either current_positions or fps must be passed.")

    if len(current_positions) == 0:
        return False

    if len(current_positions) != len(trajectory):
        warn("Mismatch between number of positioners and trajectory.")
        return False

    array_current = numpy.zeros((len(current_positions), 2), dtype=numpy.float32)
    array_trajectory = numpy.zeros((len(current_positions), 2), dtype=numpy.float32)
    for ii, pid in enumerate(current_positions):
        array_current[ii, :] = current_positions[pid]

        if pid not in trajectory:
            warn(f"Positioner {pid} is not in the trajectory.")
            return False

        alpha0 = trajectory[pid]["alpha"][0][0]
        beta0 = trajectory[pid]["beta"][0][0]

        array_trajectory[ii, :] = (alpha0, beta0)

    if numpy.allclose(array_current, array_trajectory, atol=atol):
        return True
    else:
        warn("Trajectory start and current positions do not match.")
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-06
# @Filename: version.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from coordio import __version__ as coordio_version
from kaiju import __version__ as kaiju_version

from jaeger import __version__ as jaeger_version

from . import jaeger_parser


jaeger_parser.commands.pop("version")


@jaeger_parser.command()
async def version(command, fps):
    """Returns the versions used."""

    command.info(version=jaeger_version)
    command.info(kaiju_version=kaiju_version)
    command.info(coordio_version=coordio_version)

    try:
        from fps_calibrations import get_version

        command.info(fps_calibrations_version=get_version())
    except ImportError:
        command.warning("Cannot output fps_calibrations version.")

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-10
# @Filename: configuration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import os
import pathlib
import warnings
from copy import deepcopy
from time import time

from typing import TYPE_CHECKING, Optional, Union, cast

import numpy
import pandas
from astropy.table import Table
from astropy.time import Time
from pydl.pydlutils.yanny import write_ndarray_to_yanny

from coordio import (
    ICRS,
    Field,
    FocalPlane,
    Observed,
    PositionerApogee,
    PositionerBoss,
    Site,
    Wok,
)
from coordio import __version__ as coordio_version
from coordio.defaults import INST_TO_WAVE, POSITIONER_HEIGHT, calibration
from kaiju import __version__ as kaiju_version
from sdssdb.peewee.sdss5db import opsdb, targetdb
from sdsstools.time import get_sjd

from jaeger import FPS
from jaeger import __version__ as jaeger_version
from jaeger import config, log
from jaeger.exceptions import JaegerError, TrajectoryError
from jaeger.ieb import IEB
from jaeger.kaiju import (
    decollide_in_executor,
    dump_robot_grid,
    get_path_pair_in_executor,
    get_robot_grid,
    get_snapshot_async,
    load_robot_grid,
    warn,
)
from jaeger.utils.database import connect_database
from jaeger.utils.helpers import run_in_executor

from .tools import copy_summary_file, positioner_to_wok, wok_to_positioner


if TYPE_CHECKING:
    from clu import Command

    from jaeger.actor import JaegerActor

    from .design import Design


__all__ = [
    "BaseConfiguration",
    "Configuration",
    "ManualConfiguration",
    "AssignmentData",
    "DitheredConfiguration",
]

PositionerType = Union[PositionerApogee, PositionerBoss]


def get_fibermap_table(length: int) -> tuple[numpy.ndarray, dict]:
    """Returns a stub for the FIBERMAP table and a default entry,"""

    fiber_map_data = [
        ("positionerId", numpy.int16, -999),
        ("holeId", "U7", ""),
        ("fiberType", "U10", ""),
        ("assigned", numpy.int16, 0),
        ("on_target", numpy.int16, 0),
        ("valid", numpy.int16, 0),
        ("decollided", numpy.int16, 0),
        ("xwok", numpy.float64, -999.0),
        ("ywok", numpy.float64, -999.0),
        ("zwok", numpy.float64, -999.0),
        ("xFocal", numpy.float64, -999.0),
        ("yFocal", numpy.float64, -999.0),
        ("alpha", numpy.float32, -999.0),
        ("beta", numpy.float32, -999.0),
        ("racat", numpy.float64, -999.0),
        ("deccat", numpy.float64, -999.0),
        ("pmra", numpy.float32, -999.0),
        ("pmdec", numpy.float32, -999.0),
        ("parallax", numpy.float32, -999.0),
        ("ra", numpy.float64, -999.0),
        ("dec", numpy.float64, -999.0),
        ("lambda_eff", numpy.float32, -999.0),
        ("coord_epoch", numpy.float32, -999.0),
        ("spectrographId", numpy.int16, -999),
        ("fiberId", numpy.int16, -999),
        ("mag", numpy.dtype(("<f4", (5,))), [-999.0] * 5),
        ("optical_prov", "U30", ""),
        ("bp_mag", numpy.float32, -999.0),
        ("gaia_g_mag", numpy.float32, -999.0),
        ("rp_mag", numpy.float32, -999.0),
        ("h_mag", numpy.float32, -999.0),
        ("catalogid", numpy.int64, -999),
        ("carton_to_target_pk", numpy.int64, -999),
        ("cadence", "U100", ""),
        ("firstcarton", "U100", ""),
        ("program", "U100", ""),
        ("category", "U100", ""),
        ("sdssv_boss_target0", numpy.int64, 0),
        ("sdssv_apogee_target0", numpy.int64, 0),
        ("delta_ra", numpy.float64, 0.0),
        ("delta_dec", numpy.float64, 0.0),
    ]

    names, formats, defaults = zip(*fiber_map_data)

    fibermap = numpy.empty((length,), dtype={"names": names, "formats": formats})

    # Define a default row with all set to "" or -999. depending on column data type.
    default = {}
    for i in range(len(names)):
        name = names[i]
        default[name] = defaults[i]

    return (fibermap, default)


class BaseConfiguration:
    """A base configuration class."""

    assignment_data: BaseAssignmentData
    epoch: float | None

    def __init__(self, scale: float | None = None):

        if len(calibration.positionerTable) == 0:
            raise ValueError("FPS calibrations not loaded or the array is empty.")

        # Configuration ID is None until we insert in the database.
        # Once set, it cannot be changed.
        self.configuration_id: int | None = None
        self._summary_file: str | None = None

        self.scale = scale or 1.0

        # Whether the configuration is a dither. If True, there will be a base
        # configuration from which we dithered and the trajectory will be applied
        # ignoring collisions and with early exit for deadlocks.
        self.parent_configuration: BaseConfiguration | None = None
        self.is_dither: bool = False

        self.is_cloned: bool = False
        self.cloned_from: int | None = None

        self.design: Design | None = None
        self.design_id: int | None = None

        self.extra_summary_data = {}

        self.fps = FPS.get_instance()

        self.robot_grid = self._initialise_grid()

        self.command: Command[JaegerActor] | None = None

        self._decollided: list[int] = []

        self.to_destination: dict | None = None
        self.from_destination: dict | None = None

        self.created_time = time()
        self.executed: bool = False

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k in ["robot_grid", "command"]:
                v = None
            setattr(result, k, deepcopy(v, memo))
        return result

    async def clone(
        self,
        design_id: int | None = None,
        copy_summary_F: bool = False,
        write_to_database: bool = True,
        write_summary: bool = True,
    ) -> BaseConfiguration:
        """Clones a configuration.

        Parameters
        ----------
        design_id
            If set, modifies the ``design_id`` in the configuration `.Design`
            instance and copied summary files.
        copy_summary_F
            If `True` and a `confSummaryF` exists for the current configuration,
            copies it with the newly assigned configuration_id.
        write_to_database
            Write new configuration to database
        write_summary
            Write summary file for the new configuration.

        Returns
        -------
        configuration
            The cloned configuration.

        """

        assert self.configuration_id is not None, "configuration_id not set."

        original_configuration_id = self.configuration_id

        new = self.copy()
        new.robot_grid = self.robot_grid

        new.configuration_id = None
        new.is_cloned = True
        new.cloned_from = original_configuration_id

        if design_id is not None and new.design_id is not None:
            if new.design:
                new.design.design_id = design_id
            new.design_id = design_id

        if write_to_database:
            new.write_to_database()

        if write_to_database and write_summary:
            await new.write_summary(headers={"cloned_from": new.cloned_from})

        if write_to_database and copy_summary_F:
            assert new.configuration_id is not None
            copy_summary_file(
                original_configuration_id,
                new.configuration_id,
                new.design_id,
                "F",
            )

        return new

    def copy(self):
        """Returns a deep copy of the configuration instance. Drops the robot grid."""

        return deepcopy(self)

    def set_command(self, command: Optional[Command[JaegerActor]]):
        """Sets the command to which to output messages."""

        self.command = command

    def log(self, msg: str, level=logging.INFO, to_command: bool = True):
        """Log message to the log and command."""

        log.log(level, msg)

        if to_command and self.command and self.command.status.is_done is False:
            if level == logging.DEBUG:
                self.command.debug(msg)
            elif level == logging.INFO:
                self.command.info(msg)
            elif level == logging.WARNING:
                self.command.warning(msg)
            elif level == logging.ERROR:
                self.command.error(msg)

    def _initialise_grid(self, collision_buffer: float | None = None):
        """Create a robot grid."""

        self.robot_grid = get_robot_grid(self.fps, collision_buffer=collision_buffer)

        return self.robot_grid

    def __repr__(self):
        return f"<Configuration (configuration_id={self.configuration_id}>"

    def recompute_coordinates(self, jd: Optional[float] = None):
        """Recalculates the coordinates.

        Parameters
        ----------
        jd
            The Julian Date for which to compute the coordinates.

        """

        if self.configuration_id is not None:
            raise JaegerError(
                "Cannot recompute coordinates once the configuration "
                "ID has been set."
            )

        self.assignment_data.compute_coordinates(jd=jd)

        assert self.assignment_data.site.time
        self.epoch = self.assignment_data.site.time.jd

    async def get_paths(
        self,
        collision_buffer: float | None = None,
        decollide: bool = True,
        simple_decollision: bool = False,
        resolve_deadlocks: bool = True,
        n_deadlock_retries: int = 5,
        path_generation_mode: str | None = None,
        force: bool = False,
    ) -> dict:
        """Returns a trajectory dictionary from the folded position.

        Also stores the to destination trajectory so that it can be
        used later to return to folded position.

        Parameters
        ----------
        collision_buffer
            The collision buffer to use when generating paths. If `None`, defaults
            to the configuration file value.
        decollide
            Runs the decollision routine.
        simple_decollision
            If `True`, runs `decollideGrid()` without trying to prioritise and
            minimise what robots move.
        resolve_deadlocks
            Whether to solve for deadlocks after decollision.
        n_deadlock_retries
            How many times to try solving deadlocks.
        path_generation_mode
            The path generation mode, either ``'greedy'`` or ``'mdp'``. If not
            defined, uses the default path generation mode.
        force
            If `False`, fails if the robot grid is deadlocked. Always fails if there
            are collisions.

        Returns
        -------
        from_destination
            Returns the ``from destination`` trajectory (usually from lattice to
            targets).

        """

        assert isinstance(self, BaseConfiguration)

        # Just to be sure, reinitialise the grid.
        self.robot_grid = self._initialise_grid(collision_buffer=collision_buffer)

        ftable = self.assignment_data.fibre_table
        alpha0, beta0 = config["kaiju"]["lattice_position"]

        # Assign positions to all the assigned, valid targets.
        # TODO: remove disabled from here.
        valid = ftable.loc[(ftable.assigned == 1) & (ftable.valid == 1)]

        self.log(
            f"Assigned targets {(ftable.assigned == 1).sum()}. "
            f"Valid targets {len(valid)}."
        )

        invalid = []
        for robot in self.robot_grid.robotDict.values():
            if robot.isOffline:
                ftable.loc[robot.id, "offline"] = 1
                invalid.append(robot.id)
                continue

            if robot.id not in valid.index.get_level_values(0):
                robot.setAlphaBeta(alpha0, beta0)
                robot.setDestinationAlphaBeta(alpha0, beta0)
                robot.setXYUniform()  # Scramble unassigned robots.
                invalid.append(robot.id)
            else:
                vrow = valid.loc[robot.id].iloc[0]
                robot.setAlphaBeta(vrow.alpha, vrow.beta)
                robot.setDestinationAlphaBeta(alpha0, beta0)

        self._update_coordinates(invalid)

        decollided: list[int] = []
        if decollide:
            priority_order = valid.index.get_level_values(0).tolist()

            self.robot_grid, decollided = await decollide_in_executor(
                self.robot_grid,
                simple=simple_decollision,
                priority_order=priority_order,
            )

            if len(decollided) > 0:
                self.log(
                    f"{len(decollided)} positioners were collided and "
                    f"were reassigned: {decollided}.",
                    level=logging.WARNING,
                    to_command=False,
                )

            self._update_coordinates(decollided)

            # Final check for collisions.
            if len(self.robot_grid.getCollidedRobotList()) > 0:
                raise TrajectoryError("The robot grid remains collided.")

        # Fix deadlocks (this sets the trajectories in the instance).
        self.log("Generating path pair.")
        n_retries = n_deadlock_retries if resolve_deadlocks else -1
        unlocked = await self._resolve_deadlocks(
            n_retries=n_retries,
            path_generation_mode=path_generation_mode,
            force=force,
        )

        self._update_coordinates(unlocked)

        self._decollided = list(set(decollided + unlocked))
        self.assignment_data.fibre_table.loc[self._decollided, "decollided"] = 1

        if self.from_destination is None:
            raise TrajectoryError("Cannot find valid trajectory.")

        return self.from_destination

    async def _resolve_deadlocks(
        self,
        n_retries: int = 5,
        path_generation_mode: str | None = None,
        force: bool = False,
    ) -> list[int]:
        """Iteratively fix deadlocks."""

        # Save the grid data in case we need to decollide.
        grid_data = dump_robot_grid(self.robot_grid)

        attempt: int = 0
        decollided: list[int] = []

        while True:
            result = await get_path_pair_in_executor(
                self.robot_grid,
                path_generation_mode=path_generation_mode,
            )
            self.to_destination, self.from_destination, did_fail, deadlocks = result

            n_deadlocks = len(deadlocks)

            if did_fail:
                attempt += 1

                if n_retries < 0:
                    # n_retries == -1 means we don't want to solve for deadlocks. Fail!
                    raise TrajectoryError(
                        "Failed generating a valid trajectory. "
                        f"{n_deadlocks} deadlocks were found."
                    )

                if attempt > n_retries:
                    msg = (
                        f"Attempt {attempt}: {n_deadlocks} deadlocks remain but "
                        "the number of retries has been exhausted."
                    )

                    if force is False:
                        raise TrajectoryError(msg)

                    else:
                        self.log(msg, level=logging.WARNING, to_command=False)

                # Replace one of the deadlocked robots with a random new position.
                # TODO: maybe not call setXYUniform and do a small offset.

                if attempt == 1:
                    self.log("Deadlocks found. Attempting resolution.")

                self.log(f"Attempt {attempt}: {n_deadlocks} deadlocks found.")

                to_move = numpy.random.choice(deadlocks)
                self.log(f"Trying to unlock positioner {to_move}.", level=logging.DEBUG)

                # Restore robot grid (it has been mangled by calling get_path_pair).
                self.robot_grid = load_robot_grid(grid_data)

                # Assign new position to the random deadlocked robot.
                self.robot_grid.robotDict[to_move].setXYUniform()

                # Now check if it's collided and decollide it.
                if self.robot_grid.isCollided(to_move):
                    if self.robot_grid.robotDict[to_move].isOffline is False:
                        self.robot_grid.decollideRobot(to_move)
                        if self.robot_grid.isCollided(to_move):
                            raise TrajectoryError("Cannot decollide deadlocked robot.")

                if to_move not in decollided:
                    decollided.append(to_move)

                grid_data = dump_robot_grid(self.robot_grid)

            else:
                if attempt > 1:
                    self.log("All deadlocks have been fixed.")
                break

        return decollided

    def _update_coordinates(
        self,
        positioner_ids: list[int] = [],
        mark_off_target: bool = True,
    ):
        """Updates the coordinates of a series of robots."""

        ftable = self.assignment_data.fibre_table

        n_positioners = len(positioner_ids)
        if n_positioners > 0:
            self.log(f"Recomputing {n_positioners} coordinates.", level=logging.DEBUG)

        # Update the [xyz]wok_kaiju columns with the values the kaiju uses.
        # These should (!) be identical to [xyz]wok. We do this for all the
        # positioners in the grid.
        for robot in self.robot_grid.robotDict.values():
            cols = ["xwok_kaiju", "ywok_kaiju", "zwok_kaiju"]
            ftable.loc[(robot.id, "APOGEE"), cols] = robot.apWokXYZ
            ftable.loc[(robot.id, "BOSS"), cols] = robot.bossWokXYZ
            ftable.loc[(robot.id, "Metrology"), cols] = robot.metWokXYZ

            # If the robot is in the list it means now it's at a different position
            # now, so we need to update its coordinates in the table.
            if robot.id in positioner_ids:
                for ftype in ["APOGEE", "BOSS", "Metrology"]:
                    self.assignment_data.positioner_to_icrs(
                        robot.id,
                        ftype,
                        robot.alpha,
                        robot.beta,
                        self.design.field.position_angle if self.design else 0.0,
                        update=True,
                    )

        self.assignment_data.validate()

        if mark_off_target:
            ftable.loc[positioner_ids, "on_target"] = 0

    async def save_snapshot(self, highlight=None):
        """Saves a snapshot of the current robot grid."""

        if self.assignment_data and self.assignment_data.observatory:
            observatory = self.assignment_data.observatory.upper()
        else:
            observatory = None

        mjd = get_sjd(observatory)
        dirpath = os.path.join(config["fps"]["configuration_snapshot_path"], str(mjd))
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        cid = self.configuration_id or -999

        path = os.path.join(dirpath, f"configuration_snapshot_{mjd}_{cid}.pdf")

        data = dump_robot_grid(self.robot_grid)

        title = None
        if self.configuration_id:
            title = f"Configuration {self.configuration_id}"

        await run_in_executor(
            get_snapshot_async,
            path,
            data=data,
            highlight=highlight or self._decollided,
            title=title,
            executor="process",
        )

        return path

    @property
    def ingested(self):
        """Returns `True` if the configuration has been loaded to opsdb."""

        if self.configuration_id is None:
            return False

        return (
            opsdb.Configuration.select()
            .where(opsdb.Configuration.configuration_id == self.configuration_id)
            .exists()
        )

    def write_to_database(self, replace=False):
        """Writes the configuration to the database."""

        if connect_database(targetdb.database) is False:
            raise RuntimeError("Cannot connect to database.")

        if "admin" in targetdb.database._config:
            targetdb.database.become_admin()

        assert self.assignment_data.site.time
        epoch = self.assignment_data.site.time.jd

        if self.configuration_id is None:

            with opsdb.database.atomic():

                if self.design is None:
                    if (
                        targetdb.Design()
                        .select()
                        .where(targetdb.Design.design_id == self.design_id)
                        .exists()
                    ):
                        design_id = self.design_id
                    else:
                        design_id = None
                else:
                    design_id = self.design.design_id

                configuration = opsdb.Configuration(
                    design_id=design_id,
                    epoch=epoch,
                    calibration_version=calibration.fps_calibs_version,
                )
                configuration.save()

            if configuration.configuration_id is None:
                raise JaegerError("Failed loading configuration.")

            self.configuration_id = configuration.configuration_id

        else:

            if self.configuration_id is None:
                raise JaegerError("Must have a configuration_id to replace.")

            if not self.ingested:
                raise JaegerError(
                    f"Configuration ID {self.configuration_id} does not exists "
                    "in opsdb and cannot be replaced."
                )

            if replace is False:
                raise JaegerError(
                    f"Configuration ID {self.configuration_id} has already "
                    "been loaded. Use replace=True to overwrite it."
                )

            with opsdb.database.atomic():
                opsdb.AssignmentToFocal.delete().where(
                    opsdb.AssignmentToFocal.configuration_id == self.configuration_id
                ).execute(opsdb.database)

                configuration = opsdb.Configuration.get_by_id(self.configuration_id)
                configuration.delete_instance()
                self.configuration_id = None

                return self.write_to_database(replace=False)

        a_data = self.assignment_data.fibre_table

        focals = []
        for index, data in a_data.iterrows():

            index = cast(tuple, index)
            pid = index[0]
            hole_id = data.hole_id
            try:
                if data.valid == 0:
                    raise ValueError(f"Invalid coordinate found for positioner {pid}.")

                xfocal, yfocal = data.xfocal, data.yfocal
                if xfocal == -999.0 or yfocal == -999.0:
                    xfocal = yfocal = None

            except ValueError:
                xfocal = yfocal = None

            if self.design and hole_id in self.design.target_data:
                assignment_pk = self.design.target_data[hole_id]["assignment_pk"]
            else:
                assignment_pk = None

            focals.append(
                dict(
                    assignment_pk=assignment_pk,
                    xfocal=xfocal,
                    yfocal=yfocal,
                    positioner_id=pid,
                    configuration_id=self.configuration_id,
                )
            )

        with opsdb.database.atomic():
            opsdb.AssignmentToFocal.insert_many(focals).execute(opsdb.database)

    @staticmethod
    def _get_summary_file_path(configuration_id: int, observatory: str, flavour: str):

        if configuration_id is None:
            raise JaegerError("Configuration ID not set.")

        if "SDSSCORE_DIR" not in os.environ:
            raise JaegerError("$SDSSCORE_DIR is not set. Cannot write summary file.")

        sdsscore_dir = os.environ["SDSSCORE_DIR"]
        path = os.path.join(
            sdsscore_dir,
            observatory.lower(),
            "summary_files",
            f"{int(configuration_id / 100):04d}XX",
            f"confSummary{flavour}-{configuration_id}.par",
        )

        return path

    async def write_summary(
        self,
        flavour: str = "",
        path: str | pathlib.Path | None = None,
        overwrite: bool = False,
        headers: dict = {},
        fibre_table: pandas.DataFrame | None = None,
    ):
        """Writes the confSummary file."""

        # TODO: some time may be saved by doing a single DB query and retrieving
        # all the info at once for all the assignments. Need to be careful
        # to maintain the order.

        if self.configuration_id is None:
            raise JaegerError("Configuration needs to be set and loaded to the DB.")

        adata = self.assignment_data
        fdata = fibre_table or self.assignment_data.fibre_table.copy()

        # Add fiberId
        fass = pandas.merge(
            calibration.positionerTable,
            calibration.fiberAssignments,
            left_index=True,
            right_index=True,
        ).set_index("positionerID_x")

        fdata.loc[
            (fass.index, "APOGEE"), "fiberId"
        ] = fass.APOGEEFiber.tolist()  # type:ignore

        fdata.loc[
            (fass.index, "BOSS"), "fiberId"
        ] = fass.BOSSFiber.tolist()  # type:ignore

        fdata.fillna(-999, inplace=True)

        time = Time.now()

        design = self.design

        if self.fps and isinstance(self.fps.ieb, IEB):
            temp: float = (await self.fps.ieb.read_device("T3"))[0]
        else:
            temp = -999.0

        header = {
            "configuration_id": self.configuration_id,
            "robostrategy_run": "NA",
            "fps_calibrations_version": calibration.fps_calibs_version,
            "jaeger_version": jaeger_version,
            "coordio_version": coordio_version,
            "kaiju_version": kaiju_version,
            "design_id": self.design_id,
            "field_id": -999,
            "focal_scale": adata.scale or 0.999882,
            "instruments": "BOSS APOGEE",
            "epoch": adata.site.time.jd if adata.site.time else -999,
            "obstime": time.strftime("%a %b %d %H:%M:%S %Y"),
            "MJD": get_sjd(adata.observatory.upper()),
            "observatory": adata.observatory,
            "temperature": round(temp, 1),
            "raCen": -999.0,
            "decCen": -999.0,
            "pa": -999.0,
            "is_dithered": 0,
            "parent_configuration": -999,
            "dither_radius": -999.0,
            "cloned_from": -999,
        }

        if design:
            header.update(
                {
                    "robostrategy_run": design.field.rs_run,
                    "field_id": design.field.field_id,
                    "raCen": design.field.racen,
                    "decCen": design.field.deccen,
                    "pa": design.field.position_angle,
                }
            )

        header.update(headers)
        header.update(self.extra_summary_data)

        fibermap, default = get_fibermap_table(len(fdata))

        i = 0
        for index, row_data in fdata.iterrows():

            index = cast(tuple, index)

            pid, fibre_type = index
            hole_id = row_data.hole_id

            # Start with the default row.
            row = default.copy()

            if fibre_type.upper() == "APOGEE":
                spec_id = 2
            elif fibre_type.upper() == "BOSS":
                spec_id = 1
            else:
                spec_id = -1

            # Update data that is valid for all fibres.
            row.update(
                {
                    "positionerId": pid,
                    "holeId": hole_id,
                    "fiberType": fibre_type.upper(),
                    "assigned": row_data.assigned,
                    "valid": row_data.valid,
                    "on_target": row_data.on_target,
                    "xwok": row_data.xwok,
                    "ywok": row_data.ywok,
                    "zwok": row_data.zwok,
                    "xFocal": row_data.xfocal,
                    "yFocal": row_data.yfocal,
                    "alpha": row_data.alpha,
                    "beta": row_data.beta,
                    "ra": row_data.ra_epoch,
                    "dec": row_data.dec_epoch,
                    "spectrographId": spec_id,
                    "fiberId": row_data.fiberId,
                }
            )

            # And now only the one that is associated with a target.
            if row_data.assigned == 1 and design and hole_id in design.target_data:
                target = design.target_data[hole_id]
                row.update(
                    {
                        "racat": target["ra"],
                        "deccat": target["dec"],
                        "pmra": target["pmra"] or -999.0,
                        "pmdec": target["pmdec"] or -999.0,
                        "parallax": target["parallax"] or -999.0,
                        "coord_epoch": target["epoch"] or -999.0,
                        "lambda_eff": target["lambda_eff"] or -999.0,
                        "catalogid": target["catalogid"] or -999.0,
                        "carton_to_target_pk": target["carton_to_target_pk"] or -999.0,
                        "cadence": target["cadence"] or "",
                        "firstcarton": target["carton"] or "",
                        "program": target["program"] or "",
                        "category": target["category"] or "",
                        "delta_ra": target["delta_ra"] or 0.0,
                        "delta_dec": target["delta_dec"] or 0.0,
                    }
                )

                optical_mag = [target[m] or -999.0 for m in ["g", "r", "i", "z"]]
                optical_mag = [-999.0] + optical_mag  # u band
                row.update(
                    {
                        "mag": optical_mag,
                        "optical_prov": target["optical_prov"] or "",
                        "bp_mag": target["bp"] or -999.0,
                        "gaia_g_mag": target["gaia_g"] or -999.0,
                        "rp_mag": target["rp"] or -999.0,
                        "h_mag": target["h"] or -999.0,
                    }
                )

            fibermap[i] = tuple(row.values())

            i += 1

        fibermap = Table(fibermap)
        fibermap.sort(["positionerId", "fiberType"])

        if path is None:
            path = self._get_summary_file_path(
                self.configuration_id,
                self.assignment_data.observatory,
                flavour,
            )

        if os.path.exists(path):
            if overwrite:
                warn(f"Summary file {os.path.basename(path)} exists. Overwriting it.")
                os.remove(path)
            else:
                raise JaegerError(f"Summary file {os.path.basename(path)} exists.")

        os.makedirs(os.path.dirname(path), exist_ok=True)

        write_ndarray_to_yanny(
            str(path),
            [fibermap],
            structnames=["FIBERMAP"],
            hdr=header,
            enums={"fiberType": ("FIBERTYPE", ("BOSS", "APOGEE", "METROLOGY", "NONE"))},
        )

        self._summary_file = str(path)

        return path


class Configuration(BaseConfiguration):
    """A configuration based on a target design."""

    assignment_data: AssignmentData

    def __init__(
        self,
        design: Design,
        epoch: float | None = None,
        scale: float | None = None,
    ):

        super().__init__(scale=scale)

        self.design = design
        self.design_id = design.design_id
        self.assignment_data = AssignmentData(self, epoch=epoch, scale=scale)

        assert self.assignment_data.site.time

        self.epoch = self.assignment_data.site.time.jd
        self.scale = scale

    def __repr__(self):
        return (
            f"<Configuration (configuration_id={self.configuration_id} "
            f"design_id={self.design_id})>"
        )


class DitheredConfiguration(BaseConfiguration):
    """A positioner configuration dithered from a parent configuration."""

    def __init__(
        self,
        parent_configuration: BaseConfiguration,
        radius: float,
        epoch: float | None = None,
    ):

        assert parent_configuration.design

        super().__init__(scale=parent_configuration.scale)

        # This needs to be set after the __init__ beccause __init__ sets
        # parent_configuration=None.
        self.parent_configuration = parent_configuration
        assert self.parent_configuration.design

        self.is_dither = True

        self.design = self.parent_configuration.design
        self.design_id = self.design.design_id

        self.assignment_data = AssignmentData(
            self,
            epoch=epoch or self.parent_configuration.epoch,
            compute_coordinates=False,
            scale=self.scale,
        )
        self.assignment_data.fibre_table = (
            self.parent_configuration.assignment_data.fibre_table.copy()
        )

        self.assignment_data.site.set_time(self.parent_configuration.epoch)

        icrs_bore = ICRS([[self.design.field.racen, self.design.field.deccen]])
        self.assignment_data.boresight = Observed(
            icrs_bore,
            site=self.assignment_data.site,
            wavelength=INST_TO_WAVE["GFA"],
        )

        self.radius = radius

        self.extra_summary_data = {
            "parent_configuration": self.parent_configuration.configuration_id or -999,
            "is_dithered": 1,
            "dither_radius": radius,
        }

    def compute_coordinates(self, new_positions: dict):

        assert self.design

        ftable = self.assignment_data.fibre_table

        data = {}

        for index, _ in self.assignment_data.fibre_table.iterrows():
            pid, ftype = cast(tuple, index)

            new_alpha = new_positions[pid]["alpha"]
            new_beta = new_positions[pid]["beta"]

            row_data = self.assignment_data.positioner_to_icrs(
                pid,
                ftype,
                new_alpha,
                new_beta,
                position_angle=self.design.field.position_angle,
                update=False,
            )

            data[(pid, ftype)] = row_data

        # Now do a single update of the whole fibre table.
        new_fibre_table = pandas.DataFrame.from_dict(data, orient="index")
        new_fibre_table.sort_index(inplace=True)
        new_fibre_table.index.set_names(("positioner_id", "fibre_type"), inplace=True)

        # Copy some of the coordinate columns back to the original table.
        cols = [
            "ra_epoch",
            "dec_epoch",
            "az",
            "alt",
            "xfocal",
            "yfocal",
            "xwok",
            "ywok",
            "zwok",
            "xtangent",
            "ytangent",
            "ztangent",
            "alpha",
            "beta",
        ]
        ftable.loc[new_fibre_table.index, cols] = new_fibre_table.loc[:, cols]

        for ax in ["x", "y", "z"]:
            ftable.loc[:, f"{ax}wok_measured"] = numpy.nan
            ftable.loc[:, f"{ax}wok_kaiju"] = numpy.nan

        ftable.loc[:, "on_target"] = 0

        # Reset all to valid, then validate.
        ftable.loc[:, "valid"] = 1
        self.assignment_data.validate()

    async def get_paths(self, collision_buffer: float | None = None):
        """Get trajectory paths."""

        self.robot_grid = self._initialise_grid(collision_buffer=collision_buffer)

        await self.fps.update_position()
        positions = self.fps.get_positions_dict()

        for robot in self.robot_grid.robotDict.values():
            if robot.isOffline:
                continue

            robot.setAlphaBeta(*positions[robot.id])
            new_alpha, new_beta = robot.uniformDither(self.radius)
            robot.setDestinationAlphaBeta(new_alpha, new_beta)

        (
            self.to_destination,
            self.from_destination,
            *_,
        ) = await get_path_pair_in_executor(
            self.robot_grid,
            path_generation_mode="greedy",
            ignore_did_fail=True,
            stop_if_deadlock=True,
            ignore_initial_collisions=True,
        )

        # Get the actual last points where we went. Due to deadlocks and
        # collisions these may not actually be the ones we set.
        new_positions = {}
        for positioner_id in self.robot_grid.robotDict:
            if positioner_id in self.to_destination:
                alpha = self.to_destination[positioner_id]["alpha"][-1][0]
                beta = self.to_destination[positioner_id]["beta"][-1][0]
            else:
                alpha, beta = positions[positioner_id]

            new_positions[positioner_id] = {"alpha": alpha, "beta": beta}

        self.compute_coordinates(new_positions)

        return self.to_destination


class ManualConfiguration(BaseConfiguration):
    """A configuration create manually.

    Parameters
    ----------
    target_data
        A dictionary containing the targeting information. It must be a
        mapping of hole ID to dictionary. The hole ID dictionaries must
        include one of the following pairs: ``ra_icrs`` and ``dec_icrs``,
        ``xwok`` and ywok``, or ``alpha`` and ``beta``, in order of priority.
        The remaining coordinates will be filled out using coordinate
        transformations. If ``ra_icrs/dec_icrs`` are passed, a value for
        ``epoch`` is also required. An additional key, ``fibre_type``
        must be set for each target with value ``'APOGEE'``, ``'BOSS'``,
        or ``'Metrology``.
    field_centre
        A tuple or array with the boresight coordinates as current epoch
        RA/Dec. If `None`, target coordinates must be positioner or wok.
    design_id
        A design identifier for this configuration.
    observatory
        The observatory name. If `None`, uses the value from the configuration.
    position_angle
        The position angle of the field.

    """

    assignment_data: ManualAssignmentData

    def __init__(
        self,
        target_data: dict[str, dict],
        field_centre: tuple[float, float] | numpy.ndarray | None = None,
        design_id: int = -999,
        observatory: str | None = None,
        position_angle: float = 0.0,
        scale: float | None = None,
    ):

        super().__init__(scale=scale)

        self.design = None
        self.design_id = design_id
        self.epoch = None

        if observatory is None:
            if config["observatory"] != "${OBSERATORY}":
                observatory = config["observatory"]
                assert isinstance(observatory, str)
            else:
                raise ValueError("Unknown site.")

        self.target_data = target_data
        self.assignment_data = ManualAssignmentData(
            self,
            target_data,
            observatory,
            field_centre=field_centre,
            position_angle=position_angle,
            scale=scale,
        )

    @classmethod
    def create_from_positions(cls, positions, **kwargs):
        """Create a manual configuration from robot positions.

        Parameters
        ----------
        positions
            A dictionary of positioner ID to a tuple of ``(alpha, beta)``.

        """

        positionerTable = calibration.positionerTable.reset_index()
        data = {}

        for _, row in positionerTable.iterrows():
            hole_id = row.holeID
            positioner_id = row.positionerID

            if positioner_id not in positions:
                raise ValueError(f"Values for positioner {positioner_id} not provided.")

            alpha, beta = positions[positioner_id]
            data[hole_id] = {"alpha": alpha, "beta": beta, "fibre_type": "Metrology"}

        return cls(data, **kwargs)


class BaseAssignmentData:
    """Information about the target assignment along with coordinate transformation."""

    boresight: Observed
    position_angle: float

    _columns = [
        ("positioner_id", numpy.int32, None),
        ("fibre_type", "U10", None),
        ("hole_id", "U10", ""),
        ("assigned", numpy.int8, 0),
        ("valid", numpy.int8, 1),
        ("on_target", numpy.int8, 0),
        ("disabled", numpy.int8, 0),
        ("offline", numpy.int8, 0),
        ("deadlocked", numpy.int8, 0),
        ("decollided", numpy.int8, 0),
        ("dubious", numpy.int8, 0),
        ("wavelength", numpy.float32, numpy.nan),
        ("fiberId", numpy.int32, -999),
        ("ra_icrs", numpy.float64, numpy.nan),
        ("dec_icrs", numpy.float64, numpy.nan),
        ("ra_epoch", numpy.float64, numpy.nan),
        ("dec_epoch", numpy.float64, numpy.nan),
        ("alt", numpy.float64, numpy.nan),
        ("az", numpy.float64, numpy.nan),
        ("xfocal", numpy.float64, numpy.nan),
        ("yfocal", numpy.float64, numpy.nan),
        ("xwok", numpy.float64, numpy.nan),
        ("ywok", numpy.float64, numpy.nan),
        ("zwok", numpy.float64, numpy.nan),
        ("xwok_kaiju", numpy.float64, numpy.nan),
        ("ywok_kaiju", numpy.float64, numpy.nan),
        ("zwok_kaiju", numpy.float64, numpy.nan),
        ("xwok_measured", numpy.float64, numpy.nan),
        ("ywok_measured", numpy.float64, numpy.nan),
        ("zwok_measured", numpy.float64, numpy.nan),
        ("xtangent", numpy.float64, numpy.nan),
        ("ytangent", numpy.float64, numpy.nan),
        ("ztangent", numpy.float64, numpy.nan),
        ("alpha", numpy.float64, numpy.nan),
        ("beta", numpy.float64, numpy.nan),
    ]

    def __init__(
        self,
        configuration: Configuration | ManualConfiguration | DitheredConfiguration,
        observatory: Optional[str] = None,
        scale: float | None = None,
    ):

        self.configuration = configuration

        self.design = configuration.design
        self.design_id = self.configuration.design_id

        self.observatory: str
        if observatory is None:
            if self.design is None:
                raise ValueError("Cannot determine observatory.")
            self.observatory = self.design.field.observatory.upper()
        else:
            self.observatory = observatory

        self.site = Site(self.observatory)
        self.site.set_time()

        self.scale = scale or 1.0

        positionerTable = calibration.positionerTable
        wokCoords = calibration.wokCoords
        if positionerTable is None or wokCoords is None:
            raise ValueError("FPS calibrations not loaded.")

        self.wok_data = pandas.merge(
            positionerTable.reset_index(),
            wokCoords.reset_index(),
            on="holeID",
        )
        self.wok_data.set_index("positionerID", inplace=True)

        if self.design:
            self.target_data = self.design.target_data
            self.position_angle = self.design.field.position_angle
        else:
            self.target_data = {}

        names, _, values = zip(*self._columns)
        self._defaults = {
            name: values[i] for i, name in enumerate(names) if values[i] is not None
        }

        self.fibre_table = self._create_fibre_table()

    def __repr__(self):
        return f"<{self.__class__.__name__} (design_id={self.design_id})>"

    def compute_coordinates(self, jd: Optional[float] = None):
        """Computes coordinates in different systems.

        Parameters
        ----------
        jd
            The Julian Date for which to compute the coordinates.

        """

        raise NotImplementedError("Must be overridden by subclasses.")

    def _create_fibre_table(self):
        """Creates an empty fibre table."""

        names, dtypes, defaults = zip(*self._columns)

        # Create empty dataframe with zero values. Fill out all the index data.
        npositioner = len(self.wok_data)
        base = numpy.zeros((npositioner * 3,), dtype=list(zip(names, dtypes)))

        for i in range(len(defaults)):
            if defaults[i] is None:
                continue
            base[names[i]] = defaults[i]

        i = 0
        for pid in self.wok_data.index.tolist():
            for ft in ["APOGEE", "BOSS", "Metrology"]:
                base["positioner_id"][i] = pid
                base["fibre_type"][i] = ft
                base["hole_id"][i] = self.wok_data.loc[pid].holeID
                i += 1

        fibre_table = pandas.DataFrame(base)

        fibre_table.fibre_type = fibre_table.fibre_type.astype("category")
        fibre_table.hole_id = fibre_table.hole_id.astype("string")

        fibre_table.set_index(["positioner_id", "fibre_type"], inplace=True)
        fibre_table = fibre_table.sort_index()

        return fibre_table

    def _init_from_icrs(self):
        """Loads fibre data from target data using ICRS coordinates."""

        alpha0, beta0 = config["kaiju"]["lattice_position"]

        data = {}
        for pid in self.wok_data.index:

            positioner_data = self.wok_data.loc[pid]
            hole_id = positioner_data.holeID

            target_fibre_type: str | None = None
            if hole_id in self.target_data:

                # First do the assigned fibre.
                ftype = self.target_data[hole_id]["fibre_type"].upper()
                target_fibre_type = ftype

                target_data = self.target_data[hole_id]
                positioner_data = self.icrs_to_positioner(
                    pid,
                    ftype,
                    target_data,
                    position_angle=self.position_angle,
                    update=False,
                    on_target=1,
                    assigned=1,
                )
                data[(pid, ftype)] = positioner_data

            else:

                # If a positioner does not have an assigned target, leave it folded.
                target_fibre_type = None
                positioner_data = {"alpha": alpha0, "beta": beta0}

            # Now calculate some coordinates for the other two non-assigned fibres.
            for ftype in ["APOGEE", "BOSS", "Metrology"]:
                if ftype == target_fibre_type:
                    continue

                icrs_data = self.positioner_to_icrs(
                    pid,
                    ftype,
                    positioner_data["alpha"],
                    positioner_data["beta"],
                    position_angle=self.position_angle,
                    update=False,
                )
                data[(pid, ftype)] = icrs_data

        # Now do a single update of the whole fibre table.
        self.fibre_table = pandas.DataFrame.from_dict(data, orient="index")
        self.fibre_table.sort_index(inplace=True)
        self.fibre_table.index.set_names(("positioner_id", "fibre_type"), inplace=True)

    def _init_from_wok(self):
        """Loads fibre data from target data using wok coordinates."""

        self._check_all_assigned()

        # Calculate positioner coordinates for each wok coordinate.
        for hole_id, data in self.target_data.items():

            xwok = data["xwok"]
            ywok = data["ywok"]
            zwok = data.get("zwok", POSITIONER_HEIGHT)
            fibre_type = data["fibre_type"]

            (alpha, beta), _ = wok_to_positioner(
                hole_id,
                self.site.name,
                fibre_type,
                xwok,
                ywok,
                zwok,
            )

            self.target_data[hole_id].update({"alpha": alpha, "beta": beta})

        # Now simply call _from_positioner()
        self._init_from_positioner()

        # We want to keep the original wok coordinates that now have been overridden
        # by calling _from_positioner.
        wok_coords = []
        for idx in range(len(self.fibre_table)):
            row: pandas.Series = self.fibre_table.iloc[idx]
            hole_id = row.hole_id
            assigned = row.assigned

            if assigned == 1 and hole_id in self.target_data:
                target = self.target_data[hole_id]
                wok_coords.append(
                    [
                        target["xwok"],
                        target["ywok"],
                        target.get("zwok", numpy.nan),
                    ]
                )
            else:
                wok_coords.append([numpy.nan] * 3)

        self.fibre_table[["xwok", "ywok", "zwok"]] = wok_coords

    def _init_from_positioner(self):
        """Loads fibre data from target data using positioner coordinates."""

        self._check_all_assigned()

        data = {}
        for pid in self.wok_data.index:

            hole_id = self.wok_data.at[pid, "holeID"]
            alpha = self.target_data[hole_id]["alpha"]
            beta = self.target_data[hole_id]["beta"]

            # Now calculate some coordinates for the other two non-assigned fibres.
            for ftype in ["APOGEE", "BOSS", "Metrology"]:

                assigned = self.target_data[hole_id]["fibre_type"] == ftype

                # If boresight has been set, go all the way from positioner to ICRS.
                # Otherwise just calculate tangent and wok.
                if self.boresight:
                    icrs_data = self.positioner_to_icrs(
                        pid,
                        ftype,
                        alpha,
                        beta,
                        position_angle=self.position_angle,
                        update=False,
                        assigned=int(assigned),
                        on_target=int(assigned),
                    )
                    data[(pid, ftype)] = icrs_data

                else:

                    wok, tangent = positioner_to_wok(
                        hole_id,
                        self.site.name,
                        ftype,
                        alpha,
                        beta,
                    )

                    row = self._defaults.copy()
                    row.update(
                        {
                            "hole_id": hole_id,
                            "alpha": alpha,
                            "beta": beta,
                            "xwok": wok[0],
                            "ywok": wok[1],
                            "zwok": wok[2],
                            "xtangent": tangent[0],
                            "ytangent": tangent[1],
                            "ztangent": tangent[2],
                            "assigned": int(assigned),
                            "on_target": int(assigned),
                        }
                    )
                    data[(pid, ftype)] = row

        # Now do a single update of the whole fibre table.
        new_data = pandas.DataFrame.from_dict(data, orient="index")
        self.fibre_table.loc[new_data.index, new_data.columns] = new_data
        self.fibre_table.index.set_names(("positioner_id", "fibre_type"), inplace=True)

    def _check_all_assigned(self):
        """Check that all the positioners are in ``target_data``."""

        if len(set(self.target_data) - set(self.wok_data.holeID)) > 0:
            raise ValueError("Not all the positioners have been assigned a target.")

    def validate(self):
        """Validates the fibre table."""

        alpha_beta = self.fibre_table[["alpha", "beta"]]
        na = alpha_beta.isna().any(axis=1)
        over_180 = self.fibre_table.beta > 180

        self.fibre_table.loc[na | over_180, "valid"] = 0
        self.fibre_table.loc[na | over_180, "on_target"] = 0

    def icrs_to_positioner(
        self,
        positioner_id: int,
        fibre_type: str,
        target_data: dict,
        position_angle: float = 0.0,
        update: bool = True,
        **kwargs,
    ):
        """Converts from ICRS coordinates."""

        hole_id = self.wok_data.loc[positioner_id].holeID
        wavelength = INST_TO_WAVE.get(fibre_type.capitalize(), INST_TO_WAVE["GFA"])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            ra = target_data["ra"]
            if target_data["delta_ra"] is not None:
                # delta_ra/delta_dec are in arcsec.
                cos_dec = numpy.cos(numpy.deg2rad(target_data["dec"]))
                ra += target_data["delta_ra"] / 3600.0 / cos_dec

            dec = target_data["dec"]
            if target_data["delta_dec"] is not None:
                dec += target_data["delta_dec"] / 3600.0

            pmra = target_data["pmra"]
            pmdec = target_data["pmdec"]
            parallax = target_data["parallax"]
            epoch = target_data["epoch"]

            icrs = ICRS(
                [[ra, dec]],
                pmra=numpy.nan_to_num(pmra, nan=0),
                pmdec=numpy.nan_to_num(pmdec, nan=0),
                parallax=numpy.nan_to_num(parallax),
                epoch=Time(epoch, format="jyear").jd,
            )

            assert self.site.time
            icrs_epoch = icrs.to_epoch(self.site.time.jd, site=self.site)

            observed = Observed(icrs_epoch, wavelength=wavelength, site=self.site)
            field = Field(observed, field_center=self.boresight)
            focal = FocalPlane(
                field,
                wavelength=wavelength,
                site=self.site,
                fpScale=self.scale,
            )
            wok = Wok(focal, site=self.site, obsAngle=position_angle)

            positioner, tangent = wok_to_positioner(
                hole_id,
                self.site.name,
                fibre_type,
                wok[0][0],
                wok[0][1],
                wok[0][2],
            )

        if update is False:
            row = self._defaults.copy()
        else:
            row = {}

        row.update(
            {
                "hole_id": hole_id,
                "wavelength": INST_TO_WAVE[fibre_type.capitalize()],
                "ra_icrs": icrs[0, 0],
                "dec_icrs": icrs[0, 1],
                "ra_epoch": icrs_epoch[0, 0],
                "dec_epoch": icrs_epoch[0, 1],
                "alt": observed[0, 0],
                "az": observed[0, 1],
                "xfocal": focal[0, 0],
                "yfocal": focal[0, 1],
                "xwok": wok[0, 0],
                "ywok": wok[0, 1],
                "zwok": wok[0, 2],
                "xtangent": tangent[0],
                "ytangent": tangent[1],
                "ztangent": tangent[2],
                "alpha": positioner[0],
                "beta": positioner[1],
            }
        )
        row.update(kwargs)

        if update:
            self.fibre_table.loc[
                (positioner_id, fibre_type), row.keys()
            ] = row  # type:ignore

        return row

    def positioner_to_icrs(
        self,
        positioner_id: int,
        fibre_type: str,
        alpha: float,
        beta: float,
        position_angle: float | None = None,
        update: bool = True,
        **kwargs,
    ):
        """Converts from positioner to ICRS coordinates."""

        wavelength = INST_TO_WAVE.get(fibre_type.capitalize(), INST_TO_WAVE["GFA"])

        assert self.site.time

        if position_angle is None:
            position_angle = self.design.field.position_angle if self.design else 0.0

        hole_id = self.wok_data.at[positioner_id, "holeID"]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            wok, tangent = positioner_to_wok(
                hole_id,
                self.site.name,
                fibre_type,
                alpha,
                beta,
            )

            focal = FocalPlane(
                Wok([wok], site=self.site, obsAngle=position_angle),
                wavelength=wavelength,
                site=self.site,
                fpScale=self.scale,
            )

            if self.boresight is not None:
                field = Field(focal, field_center=self.boresight)
                obs = Observed(field, site=self.site, wavelength=wavelength)
                icrs = ICRS(obs, epoch=self.site.time.jd)
            else:
                field = obs = icrs = None

        if update is False:
            row = self._defaults.copy()
        else:
            row = {}

        row.update(
            {
                "hole_id": hole_id,
                "wavelength": wavelength,
                "ra_epoch": icrs[0, 0] if icrs is not None else numpy.nan,
                "dec_epoch": icrs[0, 1] if icrs is not None else numpy.nan,
                "xfocal": focal[0, 0],
                "yfocal": focal[0, 1],
                "xwok": wok[0],
                "ywok": wok[1],
                "zwok": wok[2],
                "alpha": alpha,
                "beta": beta,
                "xtangent": tangent[0],
                "ytangent": tangent[1],
                "ztangent": tangent[2],
            }
        )
        row.update(kwargs)

        if update:
            self.fibre_table.loc[
                (positioner_id, fibre_type), row.keys()
            ] = row  # type:ignore

        return row


class AssignmentData(BaseAssignmentData):
    """Assignment data from a valid design with associated target information."""

    design: Design

    def __init__(
        self,
        configuration: Configuration | DitheredConfiguration,
        epoch: float | None = None,
        compute_coordinates: bool = True,
        scale: float | None = None,
    ):

        super().__init__(configuration, scale=scale)

        if compute_coordinates:
            self.compute_coordinates(epoch)

    def compute_coordinates(self, jd: Optional[float] = None):
        """Computes coordinates in different systems.

        Parameters
        ----------
        jd
            The Julian Date for which to compute the coordinates.
        positioner_ids
            If `None`, compute coordinates for all the entries in the fibre table.
            Otherwise, only update the entries for the positioner IDs listed.

        """

        self.site.set_time(jd)

        icrs_bore = ICRS([[self.design.field.racen, self.design.field.deccen]])
        self.boresight = Observed(
            icrs_bore,
            site=self.site,
            wavelength=INST_TO_WAVE["GFA"],
        )

        # Load fibre data using ICRS coordinates.
        self._init_from_icrs()

        # Final validation
        self.validate()


class ManualAssignmentData(BaseAssignmentData):
    """Assignment data from a manual configuration.

    Parameters
    ----------
    configuration
        The parent `.ManualConfiguration`.
    target_data
        A dictionary containing the targeting information. It must be a
        mapping of hole ID to dictionary. The hole ID dictionaries must
        include one of the following pairs: ``ra_icrs`` and ``dec_icrs``,
        ``xwok`` and ywok``, or ``alpha`` and ``beta``, in order of priority.
        The remaining coordinates will be filled out using coordinate
        transformations. If ``ra_icrs/dec_icrs`` are passed, a value for
        ``epoch`` is also required. An additional key, ``fibre_type``
        must be set for each target with value ``'APOGEE'``, ``'BOSS'``,
        or ``'Metrology``.
    observatory
        The observatory name. If `None`, uses the value from the configuration.
    field_centre
        A tuple or array with the boresight coordinates as current epoch
        RA/Dec. If `None`, target data must be positioner or wok coordinates.
    position_angle
        The position angle of the field.
    scale
        The focal plane scale factor.

    """

    boresight: Observed | None = None

    def __init__(
        self,
        configuration: ManualConfiguration,
        target_data: dict,
        observatory: str,
        field_centre: tuple[float, float] | numpy.ndarray | None = None,
        position_angle: float = 0.0,
        scale: float | None = None,
    ):

        super().__init__(configuration, observatory=observatory, scale=scale)

        self.target_data = target_data

        if field_centre is not None:
            self.field_centre = numpy.array(field_centre)
        else:
            self.field_centre = None

        self.position_angle = position_angle

        self.fibre_table = self._create_fibre_table()
        self.compute_coordinates()

    def compute_coordinates(self, jd: Optional[float] = None):
        """Computes coordinates in different systems.

        Parameters
        ----------
        jd
            The Julian Date for which to compute the coordinates.

        """

        self.site.set_time(jd)

        if self.field_centre:
            icrs_bore = ICRS([self.field_centre])
            self.boresight = Observed(
                icrs_bore,
                site=self.site,
                wavelength=INST_TO_WAVE["GFA"],
            )

        if len(self.target_data) == 0:
            raise ValueError("Target data has zero length.")

        sample_target = list(self.target_data.values())[0]

        if "fibre_type" not in sample_target:
            raise ValueError("fibre_type column missing from target data.")

        if "ra_icrs" in sample_target and "dec_icrs" in sample_target:
            if "epoch" not in sample_target:
                raise ValueError("Missing epoch information.")
            if self.boresight is None:
                raise ValueError(
                    "Creating a manual configuration from ICRS "
                    "coordinates requires defining a field centre."
                )
            self._init_from_icrs()
        elif "xwok" in sample_target and "ywok" in sample_target:
            self._init_from_wok()
        elif "alpha" in sample_target and "beta" in sample_target:
            self._init_from_positioner()
        else:
            raise ValueError("Target data does not contain the necessary columns.")

        # Final validation.
        self.validate()
